#	This file is part of SurrealServices.
#
#	SurrealServices is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; either version 2 of the License, or
#	(at your option) any later version.
#
#	SurrealServices is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with SurrealServices; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
package net;

use strict;
use Socket;
use threads::shared;
use Carp;

no strict 'refs';

*callfuncs = \&thread::callfuncs;

my $debug = 0;
my $tkn : shared = 0;
my $sjoin = 0;
our %agents : shared;
my $nl = "\015\012";
our $ircline = 0;
our $ready : shared = 0;
our @defer_join : shared;
our %defer_mode;
our @queue : shared;
our $remoteserv : shared;
our %servers : shared;
my $nickconn_lock : shared;
my $setmode_lock : shared;

our $valid_nick_re = qr/^[][a-zA-Z`\\\|{}_][][a-zA-Z0-9`\\\|{}_-]*$/;

my %tkn = (
	PRIVMSG 	=> ['PRIVMSG',	'!'],
	WHOIS 		=> ['WHOIS',	'#'],
	WHOWAS		=> ['WHOWAS',	'$'],
	USER		=> ['USER',	'%'],
	NICK		=> ['NICK',	'&'],
	SERVER		=> ['SERVER',	"\'"],
	LIST		=> ['LIST',	'('],
	TOPIC		=> ['TOPIC',	')'],
	INVITE		=> ['INVITE',	'*'],
	VERSION		=> ['VERSION',	'+'],
	QUIT		=> ['QUIT', 	','],
	SQUIT		=> ['SQUIT',	'-'],
	KILL		=> ['KILL',	'.'],
	INFO		=> ['INFO',	'/'],
	LINKS		=> ['LINKS',	'0'],
	STATS		=> ['STATS',	'2'],
	USERS		=> ['USERS',	'3'],
	ERROR		=> ['ERROR',	'5'],
	AWAY		=> ['AWAY',	'6'],
	CONNECT		=> ['CONNECT',	'7'],
	PING		=> ['PING',	'8'],
	PONG 		=> ['PONG',	'9'],
	OPER		=> ['OPER',	';'],
	PASS		=> ['PASS',	'<'],
	WALLOPS		=> ['WALLOPS',	'='],
	GLOBOPS		=> ['GLOBOPS',	']'],
	TIME		=> ['TIME',	'>'],
	NAMES		=> ['NAMES',	'?'],
	SJOIN		=> ['SJOIN',	'~'],
	NOTICE 		=> ['NOTICE',	'B'],
	JOIN		=> ['JOIN',	'C'],
	PART		=> ['PART',	'D'],
	MODE		=> ['MODE',	'G'],
	KICK		=> ['KICK',	'H'],
	USERHOST	=> ['USERHOST',	'J'],
	SQLINE		=> ['SQLINE',	'c'],
	UNSQLINE	=> ['UNSQLINE',	'd'],
	SVSNICK		=> ['SVSNICK',	'e'],
	SVSNOOP		=> ['SVSNOOP',	'f'],
	SVSKILL		=> ['SVSKILL',	'h'],
	SVS2MODE	=> ['SVS2MODE',	'v'],
	CHGHOST		=> ['CHGHOST', 	'AL'],
	CHGIDENT	=> ['CHGIDENT',	'AZ'],
	NETINFO		=> ['NETINFO',	'AO'],
	TSCTL		=> ['TSCTL',	'AW'],
	SWHOIS		=> ['SWHOIS',	'BA'],
	SVSO		=> ['SVSO',	'BB'],
	# One may note... that although there is a TKL Token
	# it does not appear to always be used.
	# Maybe b/c 2 vs 3 chars, nobody cares.
	TKL		=> ['TKL',	'BD'],
	SHUN		=> ['SHUN',	'BL'],
	# xref unreal bug 2073 and 2182
#	SVSJOIN		=> ['SVSJOIN',	'BR'], #3.2.1 def
#	SVSJOIN		=> ['SVSJOIN',	'BX'], #3.2.2 def
	SVSJOIN		=> ['SVSJOIN',	'SVSJOIN'], #I don't know which to use
	SVSPART		=> ['SVSPART',	'BT'],
#	SVSSILENCE	=> ['SVSSILENCE','BS'],
#	SVSNO		=> ['SVSNO',	'BV'],
	
	EOS		=> ['EOS',	'ES'],
	UMODE2		=> ['UMODE2',	"\|"]
);

#event::addhandler('EOS', undef(), undef(), __PACKAGE__, 'net::eos');
event::addhandler('NETINFO', undef(), undef(), __PACKAGE__, 'net::netinfo');
event::addhandler('KILL', undef(), undef(), __PACKAGE__, 'net::killed');
event::addhandler('VERSION', undef(), undef(), __PACKAGE__, 'net::version');

sub serv_connect() {
	my $remote = $config::conf{remote};
	my $port = $config::conf{port};

	if ($port =~ /\D/) { $port = getservbyname($port, 'tcp') }
	die "No port" unless $port;
	my $iaddr   = inet_aton($remote)               || die "no host: $remote";
	my $paddr   = sockaddr_in($port, $iaddr);
	my $proto   = getprotobyname('tcp');

	socket(IRC, PF_INET, SOCK_STREAM, $proto)  || die "socket: $!";
	connect(IRC, $paddr)    || die "connect: $!";

	#open IRC, "/home/remote/ircd/perlserv-tmp/sync.txt" or die $!;
	{ my $ofh = select(IRC); $|=1; select($ofh); }

	event::addhandler('SEOS', undef(), undef(), __PACKAGE__, 'net::eos');


	ircsendimm('PROTOCTL '.($tkn ? 'TOKEN ' : '').'NICKv2 UMODE2 TKLEXT VHP'.($sjoin ? ' SJOIN SJOIN2 SJ3' : ''),
		'PASS :'.$config::conf{pass},
		'SERVER '.$config::conf{local}.' 0 '.$config::conf{numeric}.' :'.$config::conf{info});

}

sub ircrecv {
	while(my $in = <IRC>) {
		last unless thread::alive();
		$in =~ s/\r|\n//sg;
		$ircline++ unless $in =~ /^(?:8|PING)/;
		$in = $ircline . ' ' . $in;
		print ">> $in\n" if $debug;
		thread::ircrecv($in);
	}
}

sub parse_line($) {
	my ($in) = @_;
	return unless $in;
	my ($nickconn, $cmd);
	
	if($in =~ /^:(\S+) (\S+)/) {
		$cmd = $2;
	}
	elsif ($in =~ /^(\S+)/) {
		$cmd = $1;
	}

	print " -- Processing $ircline\n" if $debug;
	print "cmd: $cmd\n" if $debug;

	my %cmdhash = (
		PING		=>	'PING',
		'8'		=>	'PING',

		EOS		=>	'EOS',
		ES		=>	'EOS',

		SERVER		=>	'SERVER',
		"\'"		=>	'SERVER',

		SQUIT		=>	'SQUIT',
		'-'		=>	'SQUIT',

		NETINFO		=>	'NETINFO',
		AO		=>	'NETINFO',

		PROTOCTL	=>	'PROTOCTL',

		JOIN		=>	'JOIN',
		C		=>	'JOIN',

		PART		=>	'PART',
		D		=>	'PART',

		SJOIN		=>	'SJOIN',
		'~'		=>	'SJOIN',

		MODE		=>	'MODE',
		G		=>	'MODE',

		PRIVMSG		=>	'MESSAGE',
		'!'		=>	'MESSAGE',
		NOTICE		=>	'MESSAGE',
		B		=>	'MESSAGE',

		AWAY		=>	'AWAY',
		'6'		=>	'AWAY',

		NICK		=>	'NICK',
		'&'		=>	'NICK',

		QUIT		=>	'QUIT',
		','		=>	'QUIT',

		KILL		=>	'KILL',
		'.'		=>	'KILL',

		KICK		=>	'KICK',
		H		=>	'KICK',

		CHGHOST		=>	'HOST',
		AL		=>	'HOST',
		SETHOST		=>	'HOST',
		AA		=>	'HOST',
		'302'		=>	'HOST',

		'340'		=>	'USERIP',

		CHGIDENT	=>	'IDENT',
		AZ		=>	'IDENT',
		SETIDENT	=>	'IDENT',
		AD		=>	'IDENT',

		TOPIC		=>	'TOPIC',
		"\\"		=>	'TOPIC',

		UMODE2		=>	'UMODE',
		'|'		=>	'UMODE',

		TSCTL		=>	'TSCTL',
		AW		=>	'TSCTL',

		VERSION		=>	'VERSION',
		'+'		=>	'VERSION',
	);

	sub PING($) {
		my ($event, $src, $dst, @args);
		$_[0] =~ /^(?:8|PING) :(\S+)$/;
		# ($event, $src, $dst, $args)
		return ('PING', undef, undef, $1);
	}

	sub EOS($) {
		my $event;
		$_[0] =~ /^:(\S+) (?:EOS|ES)/; # Sometimes there's extra crap on the end?
		set_server_state($1, $ircline);
		if($1 eq $remoteserv) { $event = 'SEOS' } else { $event = 'EOS' }
		print "Ok. we had EOS\n";
		return ($event, undef, undef, $1);
	}

	sub SERVER($) {
		if($_[0] =~ /^(?:SERVER|\') (\S+) (\S+) :(.*)$/) {
			$remoteserv = $1;
			set_server_state($1, 0);
			return;
		}
		elsif($_[0] =~ /^:(\S+) (?:SERVER|\') (\S+) (\d+) :(.*)$/) {
			# source, new server, hop count, description
			set_server_state($2, 0);
			return ('SERVER', undef, undef, $1, $2, $3, $4);
		}
	}

	sub SQUIT($) {
		$_[0] =~ /^(?:SQUIT|-) (\S+) :(.*)$/;
		# source, server, reason
		set_server_state($1, undef());
		return ('SQUIT', undef, undef, $1, $2, $3);
	}

	sub NETINFO($) {
		$_[0] =~ /^(?:NETINFO|AO) (\S+) (\S+) (\S+) (\S+) (\S+) (\S+) (\S+) :(.*)$/;
		return ('NETINFO', undef, undef, $1, $2, $3, $4, $5, $6, $7, $8);
	}

	sub PROTOCTL($) {
		$_[0] =~ /^PROTOCTL (.*)$/;
		return ('PROTOCTL', undef, undef, $1);
	}

	sub JOIN($) {
		$_[0] =~ /^:(\S+) (?:C|JOIN) (\S+)$/;
		return ('JOIN', undef, 1, $1, $2);
	}
	
	sub SJOIN($) {
		if ($_[0] =~ /^(?:\~|SJOIN) (\S+) (\S+) (.*)$/) {
			return ('SJOIN', undef, undef, parse_sjoin(undef, $1, $2, $3));
		}
		elsif($_[0] =~ /^:(\S+) (?:\~|SJOIN) (\S+) (\S+) (.*)$/) {
			return ('SJOIN', undef, undef, parse_sjoin($1, $2, $3, $4));
		}
	}

	sub PART($) {
		if($_[0] =~ /^:(\S+) (?:D|PART) (\S+) :(.*)$/) {
			return ('PART', undef, 1, $1, $2, $3);
		}
		elsif($_[0] =~ /^:(\S+) (?:D|PART) (\S+)$/) {
			return ('PART', undef, 1, $1, $2, undef);
		}
	}

	sub MODE($) {
		$_[0] =~ /^:(\S+) (?:G|MODE) (\S+) (\S+) (.*)(?: \d+)?$/;
		return ('MODE', undef, 1, $1, $2, $3, $4);
	}

	sub MESSAGE($) {
		my ($event, @args);
		if($_[0] =~ /^:(\S+) (?:\!|PRIVMSG) (\S+) :(.*)$/) {
			$event = 'PRIVMSG'; @args = ($1, $2, $3);
		}
		elsif($_[0] =~ /^:(\S+) (?:B|NOTICE) (\S+) :(.*)$/) {
			$event = 'NOTICE'; @args = ($1, $2, $3);
		}
		$args[1] =~ s/\@${config::conf{local}}.*//io;
		return ($event, 0, 1, @args);
	}

	sub AWAY($) {
		if($_[0] =~ /^:(\S+) (?:6|AWAY) :(.*)$/) {
			return ('AWAY', undef, undef, $1, $2);
		}
		elsif($_[0] =~ /^:(\S+) (?:6|AWAY) $/) {
			return ('BACK', undef, undef, $1);
		}
	}

	sub NICK($) {
		my ($event, @args);
		if($_[0] =~ /^:(\S+) (?:NICK|\&) (\S+) :?(\S+)$/) {
			return ('NICKCHANGE', undef, undef, $1, $2, $3);
		}
		elsif($_[0] =~ /^(?:NICK|\&) (\S+) (\S+) (\S+) (\S+) (\S+) (\S+) (\S+) (\S+) (\S+) :(.*)$/) {
			$nickconn = 1;
			nickconn_lock(); # XXX THIS IS BROKEN!
			return ('NICKCONN', undef, undef, $1, $2, $3, $4, $5, $6, $7, $8, $9, $10);
		}
	}

	sub QUIT($) {
		$_[0] =~ /^:(\S+) (?:QUIT|\,) :(.*)$/;
		return ('QUIT', 0, undef, $1, $2);
	}

	sub KILL($) {
	#:tabris KILL ProxyBotW :tabris.netadmin.SCnet.ops!tabris (test.)
	#:ProxyBotW!bopm@ircop.SCnet.ops QUIT :Killed (tabris (test.))
		$_[0] =~ /^:(\S+) (?:KILL|\.) (\S+) :(\S+) \((.*)\)$/;
		return ('KILL', 0, 1, $1, $2, $3, $4);
	}

	sub KICK($) {
	#:tabris KICK #diagnostics SurrealBot :i know you don't like this. but it's for science!
		$_[0] =~ /^:(\S+) (?:KICK|H) (\S+) (\S+) :(.*)$/;
		# source, chan, target, reason
		#$src = 0; #$dst = 2;
		return ('KICK', 0, undef, $1, $2, $3, $4);
	}

	sub HOST($) {
		if($_[0] =~ /^:(\S+) (?:CHGHOST|AL) (\S+) (\S+)$/) {
		#:Agent CHGHOST tabris tabris.netadmin.SCnet.ops
			return ('CHGHOST', 0, 1, $1, $2, $3);
			#setter, target, vhost
		}
		elsif($_[0] =~ /^:(\S+) (?:SETHOST|AA) (\S+)$/) {
		#:tabris SETHOST tabris.netadmin.SCnet.ops
			return ('CHGHOST', 0, 1, $1, $1, $2);
		}

		elsif ($_[0] =~ /^:(?:\S* )?302 (\S+) :(\S+?)\*?=[+-].*?\@(.*)/) {
		#:serebii.razorville.co.uk 302 leif :Jesture=+~Jesture00@buzz-3F604D09.sympatico.ca
			return ('CHGHOST', 0, 1, $1, $2, $3);
		}
	}


	sub USERIP($) {
		$_[0] =~ /^:(?:\S* )?340 (\S+) :(\S+?)\*?=[+-].*?\@((?:\.|\d)*)/;
		return ('USERIP', 0, 1, $1, $2, $3);
	}

	sub IDENT($) {
		if($_[0] =~ /^:(\S+) (?:CHGIDENT|AL) (\S+) (\S+)$/) {
			return ('CHGIDENT', 0 , 1, $1, $2, $3);
			#setter, target, IDENT
		}
		elsif($_[0] =~ /^:(\S+) (?:SETIDENT|AD) (\S+)$/) {
			return ('CHGIDENT', 0 , 1, $1, $1, $2);
			#setter, target, ident
		}
	}


	sub TOPIC($) {
		if($_[0] =~ /^:(\S+) (?:TOPIC|\)) (\S+) (\S+) (\S+) :(.*)$/) {
		#:tabris TOPIC #the_lounge tabris 1089336598 :Small Channel in search of Strong Founder for long term relationship, growth, and great conversation.
			return ('TOPIC', 0, 1, $1, $2, $3, $4, $5);
		}
		elsif($_[0] =~ /^(?:TOPIC|\)) (\S+) (\S+) (\S+) :(.*)$/) {
		# src, channel, setter, timestamp, topic
			return ('TOPIC', 0, 1, undef, $1, $2, $3, $4);
		}
	}

	sub UMODE($) {
	#:tabris | +oghaANWt
		$_[0] =~ /^:(\S+) (?:UMODE2|\|) (\S+)$/;
		# src, umodes
		# a note, not all umodes are passed
		# +s, +O, and +t are not passed. possibly others
		# also not all umodes do we care about.
		# umodes we need care about:
		# oper modes: hoaACN,O oper-only modes: HSq
		# regular modes: rxB,izV (V is only somewhat, as the ircd
		# does the conversions from NOTICE to PRIVSMG for us).

		# Yes, I'm changing the event type on this
		# It's better called UMODE, and easily emulated
		# on IRCds with only MODE.
		return ('UMODE', 0, 0, $1, $2);
	}

	sub TSCTL($) {
		$_[0] =~ /^:(\S+) (?:TSCTL|AW) alltime$/;
		ircsend(":$config::conf{local} NOTICE $1 *** Server=$config::conf{local} TSTime=".
			time." time()=".time." TSOffset=0");
		return;
	}

	sub VERSION($) {
		$_[0] =~ /^:(\S+) (?:VERSION|\+).*$/;
		return ('VERSION', 0, undef, $1);
	}

	#return unless defined($cmdhash{$cmd});
	unless (defined($cmdhash{$cmd})) {
		print "Bailing out from $ircline:$cmd for lack of cmdhash\n" if $debug;
		return;
	}
	my ($event, $src, $dst, @args) = &{$cmdhash{$cmd}}($in);
	unless (defined($event)) {
		print "Bailing out from $ircline:$cmd for lack of event\n" if $debug;
		return;
	}
	#return unless defined $event;

	my (@recipients, @out);
	if(defined($dst)) {
		#$args[$dst] = lc $args[$dst];
		@recipients = split(/\,/, $args[$dst]);
	}
	#if(defined($src)) { $args[$src] = lc $args[$src]; }

	if(@recipients > 1) {
		foreach my $rcpt (@recipients) {
			$args[$dst] = $rcpt;
			push @out, [$event, $src, $dst, [@args]];
		}
	} else {
		@out = [$event, $src, $dst, [@args]];
	}

	return @out;
}

# Helper Functions

sub nickconn_lock() {
	lock $nickconn_lock;
	$nickconn_lock++;
}

sub nickconn_unlock() {
	lock $nickconn_lock;
	$nickconn_lock--;
	cond_broadcast $nickconn_lock unless $nickconn_lock;
}

sub nickconn_wait() {
	while(1) {
		lock $nickconn_lock;
		return unless $nickconn_lock;
		cond_wait $nickconn_lock;
	}
}

sub synced() {
	#return ($sync and $sync < $net::ircline);
	foreach my $s (keys(%net::servers)) {
		my $state = net::get_server_state($s);
		if(defined($state) and (!$state or $state > $net::ircline)) {
			return 0;
		}
	}

	return 1;
}

sub set_server_state($$) {
	my ($server, $state) = @_;

	lock %servers;
	if($servers{$server}) {
		return if($ircline < $servers{$server}[0]);
		$servers{$server}[0] = $ircline;
		$servers{$server}[1] = $state;
	} else {
		$servers{$server} = &share( [] );
		$servers{$server}[0] = $ircline;
		$servers{$server}[1] = $state;
	}
}

sub get_server_state($) {
	my ($server) = @_;

	my $badserver = $config::conf{'unsyncserver'};
	return 1 if($badserver and lc $server eq lc $badserver); # I HATE NEOSTATS
	
	lock %servers;
	if($servers{$server}) {
		return $servers{$server}[1];
	} else {
		return undef;
	}
}

sub is_agent_in_chan($$) {
	my ($agent, $chan) = @_;
	$agent = lc $agent; $chan = lc $chan;

	lock %agents;

	if($agents{$agent} and $agents{$agent}{CHANS} and $agents{$agent}{CHANS}{$chan}) {
		return 1;
	} else {
		return 0;
	}
}

sub agent_connect($$$$$) {
	my ($nick, $ident, $host, $modes, $gecos) = @_;
	my $time = time();

	{
		lock %agents;
	
		$agents{lc $nick} = &share( {} );
		$agents{lc $nick}{CHANS} = &share( {} );
		$agents{lc $nick}{PARMS} = &share( [] );
		for(my $i; $i < @_; $i++) {
			$agents{lc $nick}{PARMS}[$i] = $_[$i];
		}
	}
	
	$host = $config::conf{local} unless $host;
	ircsend($tkn{NICK}[$tkn]." $nick 1 $time $ident $host $config::conf{local} 1 $modes * :$gecos");
}

sub agent_quit($$) {
	my ($nick, $msg) = @_;

	{
		lock %agents;

		delete($agents{lc $nick}{CHANS});
		delete($agents{lc $nick});
	}

	ircsendimm(":$nick ".$tkn{QUIT}[$tkn]." :$msg");
}

sub agent_quit_all($) {
	my ($msg) = @_;
	
	my @agents;
	{
		lock %agents;
		@agents = keys(%agents);
	}

	foreach my $a (@agents) {
		agent_quit($a, $msg);
	}
}

sub join($$) {
	my ($agent, $chan) = @_;

	lock %agents;

	if($agents{lc $agent}) {
		lock %agents;
		$agents{lc $agent}{CHANS}{lc $chan} = 1;
		net::ircsend(":$agent ".$tkn{JOIN}[$tkn]." $chan");
	} else {
		if($ready) {
			privmsg('ServServ', $config::conf{diag}, "Tried to make nonexistent agent ($agent) join channel ($chan)");
		} else {
			print "Deferred join: $agent $chan\n";
			push @defer_join, "$agent $chan";
		}
	}
}

sub part($$$) {
	my ($agent, $chan, $reason) = @_;
	{
		lock %agents;
		delete($agents{lc $agent}{CHANS}{lc $chan});
	}
	net::ircsend(":$agent $tkn{PART}[$tkn] $chan :$reason");
}

# Handler functions

sub pong($$$) {
        my ($src, $cookie, $dst) = @_;
	# This will only make sense if you remember that
	# $src is where it came from, $dst is where it went (us)
	# we're basically bouncing it back, but changing from PING to PONG.
	if (defined($dst) and defined($cookie)) {
		# $dst is always $config::conf{local} anyway...
		# this is only valid b/c we never have messages routed THROUGH us
		# we are always an end point.
		ircsendimm(":$dst $tkn{PONG}[$tkn] $src :$cookie");
	}
	else {
		ircsendimm("$tkn{PONG}[$tkn] :$src");
        }
}

sub eos {
	print "GOT EOS\n\n" if $debug;

	#foreach my $k (keys %servers) {
	#	print "Server: $k ircline: ",$servers{$k}[0], " state: ", $servers{$k}[1], "\n";
	#}
	#print "Synced: ", synced(), "\n\n";
	#exit;
	
	ircsendimm(':'.$config::conf{local}.' '.$tkn{EOS}[$tkn]);
	#$ready = 1;
	foreach my $j (@defer_join) {
		print "Processing join: $j\n";
		my ($agent, $chan) = split(/ /, $j);
		net::join($agent, $chan);
	}
	undef(@defer_join);
	
	ircsendimm(@queue) if @queue;
	$ready = 1;
}

sub netinfo($$$$$$$$) {
	ircsendimm($tkn{NETINFO}[$tkn].' 0 '.time." $_[2] $_[3] 0 0 0 :$_[7]");
	$config::conf{network} = $_[7];
}

sub killed($$$$) {
	my ($src, $dst, $path, $reason) = @_;
	lock %agents;
	if (defined($agents{lc $dst})) {
		if($src =~ /\./) {
			irckill($config::conf{'local'}, $dst, "Nick Collision");
		} else {
			irckill('ServServ', $src, "Do not kill services agents.");
		}
		
		&agent_connect($agents{lc $dst}{PARMS});
	}
}

sub tssync {
	ircsendimm(":$config::conf{local} $tkn{TSCTL}[$tkn] SVSTIME ".time);
}

sub parse_sjoin($$$$) {
	my ($server, $ts, $cn, $parms) = @_;
	my (@users, @bans, @excepts, @blobs, $blobs, $chmodes, $chmodeparms);
	
	$server = '' unless $server;

	if($parms =~ /^:(.*)/) {
		$blobs = $1;
	} else {
		($chmodes, $blobs) = split(/ :/, $parms, 2);
		($chmodes, $chmodeparms) = split(/ /, $chmodes, 2);
	}
	@blobs = split(/ /, $blobs);
	
	foreach my $x (@blobs) {
		if($x =~ /^(\&|\")(.*)$/) {
			my $type;
			push @bans, $2 if $1 eq '&';
			push @excepts, $2 if $1 eq '"';
		} else {
			$x =~ /^([*~@%+]*)(.*)$/;
			my ($prefixes, $nick) = ($1, $2);
			my @prefixes = split(//, $prefixes);
			my $op;
			foreach my $prefix (@prefixes) {
				$op |= $modes::modes{q} if ($prefix eq '*');
				$op |= $modes::modes{a} if ($prefix eq '~');
				$op |= $modes::modes{o} if ($prefix eq '@');
				$op |= $modes::modes{h} if ($prefix eq '%');
				$op |= $modes::modes{v} if ($prefix eq '+');
			}

			push @users, { NICK => $nick, __OP => $op };
		}
	}
	
	return ($server, $cn, $ts, $chmodes, $chmodeparms, \@users, \@bans, \@excepts);
}

# Send Functions

sub ircsendreal {
	foreach my $line (@_) {
		print IRC "$line\r\n";
		print "<< $line\n" if $debug;
	}
}

sub ircsend {
	#print "IRCSEND: ready:$ready shutdown:$thread::shutdown\n\n";
	return unless @_;

	if($debug) {
		foreach my $x (@_) {
			print "<< $ircline $x\n";
		}
	}

	unless(thread::alive()) { ircsendreal(@_); return; }
	if($ready) {
		thread::ircsend(@_);
	} else {
		foreach my $x (@_) {
			if($x =~ /^$tkn{NICK}[$tkn]/) {
				unshift @queue, $x;
			} else {
				push @queue, $x;
			}
		}
	}
}

sub ircsendimm {
	thread::ircsend(@_);
}

sub kick($$$$) {
	my ($src, $chan, $target, $reason) = @_;
	ircsend(":$src $tkn{KICK}[$tkn] $chan $target :$reason");
#	thread::ircrecv(":$src $tkn{KICK}[$tkn] $chan $target :$reason");
	callfuncs('KICK', 0, 2, [$src, $chan, $target, $reason]);
}

sub invite($$$) {
	my ($src, $chan, $target) = @_;
	#:SecurityBot INVITE tabris #channel
	ircsend(":$src $tkn{INVITE}[$tkn] $target $chan");
}

sub ping {
#	if(@_ == 1) {
		ircsend(':'.$config::conf{local}.' '.$tkn{PING}[$tkn].' :'.$config::conf{local});
#	} else {
#		ircsend(':'.$_[2].' '.$tkn{PONG}[$tkn].' '.$_[0].' :'.$_[1]);
#	}
}

sub privmsg($$@) {
	my ($src, $dst, @msgs) = @_;

	if ($debug) {
	foreach my $buf (@_) {
		# 3 spaces, two colons, PRIVMSG=7
		my $buflen = length($buf) + length($src) + length($dst) + 12;
		net::privmsg($config::conf{local}, $config::conf{diag},
			"\002BUG BUG BUG BUG\002 buflen: $buflen")
			if ($buflen > 508);
	}
	}

	# submit a list of messages as a single packet to the server
	ircsend(":$src $tkn{PRIVMSG}[$tkn] $dst :".join("\r\n".":$src $tkn{PRIVMSG}[$tkn] $dst :", @msgs));
}

sub debug_privmsg($$@) {
	my ($src, $dst, @msgs) = @_;

	if ($debug) {
	foreach my $buf (@_) {
		# 3 spaces, two colons, PRIVMSG=7
		my $buflen = length($buf) + length($src) + length($dst) + 12;
		net::debug_privmsg($config::conf{local}, $config::conf{diag},
			"\002BUG BUG BUG BUG\002 buflen: $buflen")
			if ($buflen > 508);
	}
	}

	# submit a list of messages as a single packet to the server
	ircsendimm(":$src $tkn{PRIVMSG}[$tkn] $dst :".join("\r\n".":$src $tkn{PRIVMSG}[$tkn] $dst :", @msgs));
}

sub notice($$@) {
	my ($src, $dst, @msgs) = @_;

	if ($debug) {
	foreach my $buf (@_) {
		# 3 spaces, two colons, NOTICE=6
		my $buflen = length($buf) + length($src) + length($dst) + 11;
		net::privmsg($config::conf{local}, $config::conf{diag},
			"\002BUG BUG BUG BUG\002 buflen: $buflen")
			if ($buflen > 508);
	}
	}

	# submit a list of notices as a single packet to the server
	ircsend(":$src $tkn{NOTICE}[$tkn] $dst :".join("\r\n".":$src $tkn{NOTICE}[$tkn] $dst :", @msgs));
}

sub setumode($$$) {
	my ($src, $dst, $modes) = @_;

	ircsend(":$src $tkn{SVS2MODE}[$tkn] $dst $modes");
	callfuncs('UMODE2', 0, undef, [$dst, $modes]);
}

sub setagent_umode($$) {
	my ($src, $modes) = @_;

	ircsend(":$src $tkn{UMODE2}[$tkn] $modes");
}

sub setmode($$$;$) {
	my ($src, $dst, $modes, $parms) = @_;

	callfuncs('MODE', undef, 1, [$src, $dst, $modes, $parms]);
	
	print "$ircline -- setmode($src, $dst, $modes, $parms)\n" if $debug;
	my $prev = $defer_mode{"$src $dst"}[-1];

	if(defined($prev) and length($prev) < 300) {
		$defer_mode{"$src $dst"}[-1] = modes::merge(
			$prev, "$modes $parms", ($dst =~ /^#/ ? 1 : 0));
		print $defer_mode{"$src $dst"}[-1], " *** \n" if $debug;
	} else {
		push @{$defer_mode{"$src $dst"}}, "$modes $parms";
	}
}

sub flushmodes {
	my @k = keys(%defer_mode); my @v = values(%defer_mode);
	
	for(my $i; $i<@k; $i++) {
		my ($src, $dst) = split(/ /, $k[$i]);
		my @m = @{$v[$i]};
		foreach my $m (@m) {
			my ($modes, $parms) = split(/ /, $m, 2);

			setmode_real($src, $dst, $modes, $parms);
		}
	}

	%defer_mode = ();
}

sub setmode_real($$$;$) {
	my ($src, $dst, $modes, $parms) = @_;

	print "$ircline -- setmode_real($src, $dst, $modes, $parms)\n" if $debug;
	ircsend(":$src ".$tkn{MODE}[$tkn]." $dst $modes".($parms?" $parms":''));
}

sub settopic($$$$$) {
	my ($src, $chan, $setter, $time, $topic) = @_;
	
	ircsend(":$src ".$tkn{TOPIC}[$tkn]." $chan $setter $time :$topic");
	callfuncs('TOPIC', undef, undef, [$src, $chan, $setter, $time, $topic]);
}

sub wallops ($$) {
	my ($src, $message) = @_;
	ircsend(":$src $tkn{WALLOPS}[$tkn] :$message");
}

sub globops ($$) {
	my ($src, $message) = @_;
	ircsend(":$src $tkn{GLOBOPS}[$tkn] :$message");
}

sub kline ($$$$$) {
        my ($setter, $ident, $host, $expiry, $reason) = @_;
	$setter=$config::conf{local} unless defined($setter);
	$ident = '*' unless defined($ident);


	#foreach my $ex (@except) { return 1 if $mask =~ /\Q$ex\E/i; }
	
	#my $line = "GLINE $mask $time :$reason";
	# you need to use TKL for this. GLINE is a user command
	# TKL is a server command.	
        # format is
        # TKL +/- type ident host expiretime timenow :reason
#:nascent.surrealchat.net TKL + G * *.testing.only tabris!northman@netadmin.SCnet.ops 1089168439 1089168434 :This is just a test.
        my $line = "TKL + G $ident $host $setter ".($expiry + time()).' '.time()." :$reason";

	#notice($nick, $conf{"Debugger"}, $line) if $conf{"Debug"};
	ircsend($line);
}

sub unkline ($$$) {
	my ($setter, $ident, $host) = @_;
# TKL - G ident *.test.dom tabris!northman@netadmin.SCnet.ops
	ircsend("TKL - G $ident $host $setter");
}

sub zline ($$$$) {
        my ($setter, $host, $expiry, $reason) = @_;
	$setter=$config::conf{local} unless defined($setter);

	#foreach my $ex (@except) { return 1 if $mask =~ /\Q$ex\E/i; }
	
	#my $line = "GZLINE $mask $time :$reason";
	# you need to use TKL for this. GLINE is a user command
	# TKL is a server command.	
        # format is
        # TKL +/- type ident host expiretime timenow :reason
        my $line = "TKL + Z * $host $setter ".($expiry + time).' '.time." :$reason";
	#notice($nick, $conf{"Debugger"}, $line) if $conf{"Debug"};
	ircsend($line);
}

sub unzline ($$$) {
	my ($setter, $ident, $host) = @_;
# TKL - G ident *.test.dom tabris!northman@netadmin.SCnet.ops
	ircsend("TKL - Z $ident $host $setter");
}

sub irckill($$$) {
	my ($src, $target, $reason) = @_;
	#FIXME: this needs to be checked for correctness
	ircsendimm(":$src ".$tkn{KILL}[$tkn]." $target :$src ($reason)");
	callfuncs('KILL', 0, 1, [$src, $target, $src, $reason]);
}

sub svssno($$$) {
    my ($src, $target, $snomasks) = @_;
    $src=$config::conf{local} unless defined($src);
    # TODO:
    # None, this doesn't affect us.

    # SVSSNO is not in tokens.txt nor msg.h
    # however there is SENDSNO. no clue which to use.
    # SENDSNO may be a command by which snomasks are 
    # passed from server to server.
    # not observed in the wild yet.
    ircsend(":$src ".'SVS2SNO'." $target $snomasks ".time);
}

sub svsnick($$$) {
    my ($src, $oldnick, $newnick) = @_;
    $src=$config::conf{local} unless defined($src);
    # note: we will get a NICK cmd back after a 
    # successful nick change.
    # warning, if misused, this can KILL the user
    # with a collision
    
#    ircsend(":$src ".$tkn{SVSNICK}[$tkn]." $oldnick $newnick ".time);
    ircsend($tkn{SVSNICK}[$tkn]." $oldnick $newnick :".time);
}

sub svsnoop($$$) {
    my ($targetserver, $bool, $src) = @_;
    $src = $config::conf{local} unless defined($src);
    if ($bool > 0) { $bool = '+'; } else { $bool = '-'; }
#this is SVS NO-OP not SVS SNOOP
    ircsend(":$config::conf{local} $tkn{SVSNOOP}[$tkn] $targetserver $bool");
}

sub svswatch ($$@) {
# Changes the WATCH list of a user.
# Syntax: SVSWATCH <nick> :<watch parameters>
# Example: SVSWATCH Blah :+Blih!*@* -Bluh!*@* +Bleh!*@*.com
# *** We do not track this info nor care.
    my ($src, $target, @watchlist) = @_;
    # SVSWATCH is not in msg.h
    ircsend(":$src SVSWATCH $target :".join(' ', @watchlist));
}

sub svssilence ($$@) {
# Changes the SILENCE list of a user.
# Syntax: SVSSILENCE <nick> :<silence parameters>
# Example: SVSSILENCE Blah :+Blih!*@* -Bluh!*@* +Bleh!*@*.com
# *** We do not track this info nor care.
    my ($src, $target, @silencelist) = @_;
    # SVSSILENCE is not in msg.h
    ircsend(":$src SVSSILENCE $target :".join(' ', @silencelist));
}

sub svso($$$) {
# Gives nick Operflags like the ones in O:lines.
# SVSO <nick> <+operflags> (Adds the Operflags)
# SVSO <nick> - (Removes all O:Line flags)
# Example: SVSO SomeNick +bBkK
# *** We do not track this info nor care.
# *** We will see any umode changes later.
# *** this cmd does not change any umodes!

    my ($src, $target, $oflags) = @_;
    $src = $config::conf{local} unless defined($src);
    ircsend(":$src $tkn{SVSO}[$tkn] $target $oflags");

}

sub swhois($$$) {
# *** We do not track this info nor care.
    my ($src, $target, $swhois) = @_;
    $src = $config::conf{local} unless defined($src);
    ircsend(":$src $tkn{SWHOIS}[$tkn] $target :$swhois");
}

sub svsjoin($$@) {
    my ($src, $target, @chans) = @_;
    # a note. a JOIN is returned back to us on success
    # so no need to process this command.
    # similar for svspart.
    ircsend(($src?":$src":'')." $tkn{SVSJOIN}[$tkn] $target ".join(',', @chans));
}

sub svspart($$@) {
    my ($src, $target, @chans) = @_;
    ircsend(($src?":$src":'')." $tkn{SVSPART}[$tkn] $target ".join(',', @chans));
}

sub sqline ($;$) {
# possibly implement with TKL Q?
# we need to sqline most/all of our agents.
# tho whether we want to put it in agent_connect
# or leave it to the module to call it...
    my ($nickmask, $reason) = @_;
    ircsend("$tkn{SQLINE}[$tkn] $nickmask".($reason?" :$reason":''));
}

sub qline($$$) {
	my ($nickmask, $expiry, $reason) = @_;
# TKL version - Allows timed qlines.
# TKL + Q * test services.SC.net 0 1092179497 :test
    ircsend('TKL + Q * '.$nickmask.' '.$config::conf{local}.' '.($expiry ? ($expiry + time()) : 0).' '.time().' '.$reason);
}

sub unsqline ($) {
# possibly implement with TKL Q?
# we need to sqline most/all of our agents.
# tho whether we want to put it in agent_connect
# or leave it to the module to call it...
    my ($nickmask) = @_;
    ircsend($tkn{UNSQLINE}[$tkn].' '.$nickmask);
}

sub unqline($) {
	my ($nickmask) = @_;
# TKL version
# TKL - Q * test services.SC.net

    ircsend('TKL - Q * '.$nickmask.' '.$config::conf{local});
}

sub svskill($$$) {
	my ($src, $target, $reason) = @_;
	# SVSKILL requires a src, it will NOT work w/o one.
	# not sure if it'll accept a servername or not.
	# consider defaulting to ServServ
	die('svskill called w/o $src') unless $src;
	ircsend(':'.$src.' '.$tkn{SVSKILL}[$tkn].' '.$target.' :'.$reason);
	callfuncs('QUIT', 0, undef, [$target, $reason]);
}

sub version($) {
	my ($src) = @_;
	ircsend(":$config::conf{local} 351 $src $main::progname ver $main::version $config::conf{local} ".
		$main::extraversion);
}

sub userhost($) {
	my ($target) = @_;
	ircsend($tkn{USERHOST}[$tkn]." $target");
}

sub userip($) {
	my ($target) = @_;
	ircsend(":ServServ USERIP $target");
}

sub chghost($$$) {
	my ($src, $target, $vhost) = @_;
	ircsend(($src?":$src ":'').$tkn{CHGHOST}[$tkn]." $target $vhost");
        callfuncs('CHGHOST', 0, 1, [$src, $target, $vhost]);
}

sub chgident($$$) {
	my ($src, $target, $ident) = @_;
	ircsend(($src?":$src ":'').$tkn{CHGIDENT}[$tkn]." $target $ident");
        callfuncs('CHGIDENT', 0, 1, [$src, $target, $ident]);
}

1;
