#	This file is part of SurrealServices.
#
#	SurrealServices is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; either version 2 of the License, or
#	(at your option) any later version.
#
#	SurrealServices is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with SurrealServices; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
package thread;

use strict;
no strict 'refs';
use threads;
use threads::shared;
use FreezeThaw qw(freeze thaw);

my $debug = 0;

our $workers : shared = 0;
my @hdlrs;
my $worker;

our @inqueue : shared;
our @outqueue : shared;
our $timercond : shared;

sub spawn() {
	module::load();
	my (@workers, $timer);

	for(my $i; $i < $config::conf{'procs'}; $i++) {
		push @workers, threads->new(\&thread::worker);
		print scalar(@workers), " threads spawned\n";
	}

	$timer = threads->new(\&event::checktimers);

	foreach my $t (@workers) { $t->join; }
	$timer->join;

	module::unload_lazy();
	exit();
}

sub wait_for_worker() {
	while(1) {
		lock $workers;
		last if $workers;
		cond_wait($workers);
		last if $workers;
	}
}

sub shutdown() {
	net::agent_quit_all("Shutting down.");
	
	sleep 1;

	$main::status = main::ST_SHUTDOWN();
	
	{
		lock @thread::inqueue;
		cond_broadcast(@thread::inqueue);
	}

	{
		lock @thread::outqueue;
		cond_broadcast(@thread::outqueue);
	}

	{
		lock $thread::timercond;
		cond_broadcast($thread::timercond);
	}
}

sub alive() {
	if($main::status == main::ST_PRECONNECT() or $main::status == main::ST_NORMAL()) {
		return 1;
	}
	else {
		return 0;
	}
}

sub ircrecv($) {
	my $line = shift;
	my ($ircline, $line) = split(/ /, $line, 2);
	my @x = ($ircline, net::parse_line($line));
	{
		lock @inqueue;
		if($x[1][0] eq 'NICKCONN') {
			unshift @inqueue, freeze(@x);
		} else {
			push @inqueue, freeze(@x);
		}
		
		cond_signal(@inqueue);
	}
}

#sub callfuncs($$$$;$) {
sub callfuncs(@) {
	my ($event, $src, $dst, $args, $ircline) = @_;
	my @refs = @{$event::hdlr{$event}}; my @args = @$args;
	
	CFOUT: foreach my $r (@refs) {
		foreach my $hdlr (@hdlrs) {
			next CFOUT if($r->[3] eq $hdlr);
		}
		push @hdlrs, $r->[3];
			
		if(
			(!$r->[0] or lc($r->[0]) eq lc($args[$src])) and
			(!$r->[1] or lc($r->[1]) eq lc($args[$dst]) or (
				$r->[1]=~/^#$/i and $args[$dst]=~/^#/i)
			)
		) {
			enter_module($ircline, $r->[3], @args);
		}
		
		pop @hdlrs;
	}
}

sub enter_module(@) {
	my ($ircline, $sub, @args) = @_;
	
	# If this is a loopback, $ircline = 0
	$net::ircline = $ircline if($ircline);

	print "enter_module $net::ircline $sub (".join(', ', @args). ")\n";
	
	eval {
		local $SIG{__WARN__} = sub {
			privmsg($config::conf{'local'}, $config::conf{'diag'}, " -- Warning: ".$_[0]);
		};

		local $SIG{__DIE__} = sub {
			privmsg($config::conf{'local'}, $config::conf{'diag'}, " --", "-- DIED: ".$_[0], split(/\n/, Carp::longmess($@)), " --");
		};
		
		&{$sub}(@args);
	};

	net::flushmodes() if($ircline);
}

sub ircsend(@) {
	print "Thread ".threads->self->tid()." generated output.\n" if $debug;
	my $out = join("\r\n", @_);

	{
		lock @outqueue;
		push @outqueue, $out;
		cond_signal(@outqueue);
	}
}

sub worker() {
	$worker = 1;
	module::begin();

	{
		lock $workers;
		$workers++;
		cond_broadcast($workers);
	}
	
	INQUEUE: while(alive()) {
		my $line;
		while(!defined($line)) {
			lock @inqueue;
			
			unless(@inqueue) {
				cond_wait(@inqueue);
				last INQUEUE unless(alive());
			}
			
			$line = shift @inqueue;
		}
		
		print "Thread ".threads->self->tid()." got input.\n" if $debug;

		my ($ircline, @x) = thaw($line);
		foreach my $x (@x) {
			callfuncs(@$x, $ircline);
		}
	}

	{
		lock $workers;
		$workers--;
		cond_broadcast($workers);
	}

	module::end();
	
	print "Worker thread dying!\n";
}

sub writer() {
	OUTQUEUE: while(1) {
		my $line;
		
		while(!defined($line)) {
			lock @outqueue;
	
			unless(@outqueue) {
				last OUTQUEUE unless alive();
				cond_wait(@outqueue);
				last OUTQUEUE unless alive();
			}
	
			$line = shift @outqueue;
		}
	
		net::ircsendreal($line);
	}

	$main::status = main::ST_CLOSED();
}

1;
