#	This file is part of SurrealServices.
#
#	SurrealServices is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; either version 2 of the License, or
#	(at your option) any later version.
#
#	SurrealServices is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with SurrealServices; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
package nickserv;

#require Exporter;
#@ISA = qw(Exporter);
#@EXPORT_OK = qw(get_user_id get_user_nick is_identified notice);

use strict;
use threads::shared;
use Time::Local;

use constant {
	MAX_LIM => 16777215,

	F_HIDEMAIL => 1,
	F_NOMEMO => 2,
	F_NOACC => 4,
	F_NEVEROP => 8,
	F_AUTH => 16,
	F_HOLD => 32,
	F_FREEZE => 64,

	UF_FINISHED => 1,

	OLD_USER_AGE => 300,
};

*synced = \&net::synced;

our $user_table : shared;
my %enforcers : shared;
my %preenforce : shared;
our %uidcache : shared;
my %newuser : shared;
my %olduser : shared;

my $nsnick = 'NickServ';

my $cur_lock;
my $cnt_lock = 0;

my @protect_short = ('none', 'normal', 'high', 'kill');
my @protect_long = (
	'You will not be required to identify to use this nick.',
	'You must identify within 60 seconds to use this nick.',
	'You must identify before using this nick.',
	'You must identify before using this nick or you will be disconnected.'
);
	
my $err_deny = 'Permission denied.';
my $err_email = 'Your email address looks funny.';
my $err_pass = 'Invalid password.';

our (
	$nick_check, $nick_create, $nick_create_old, $nick_change, $nick_quit, $nick_delete, $nick_id_delete, $get_quit_empty_chans, $nick_chan_delete,
	$get_hostless_nicks,

	$get_nickchg, $del_nickchg_id, $add_nickchg, $reap_nickchg,
	
	$get_user_id, $get_user_nick,
	$get_nick_inval, $inc_nick_inval,
	$get_id_nicks, $get_using_nicks, $is_identified, $is_registered,
	$is_alias_of,

	$get_guest, $set_guest,

	$get_lock, $release_lock,
	
	$get_umodes, $set_umodes,
	
	$get_info, $get_host, $get_vhost, $set_vhost, $set_ident, $set_ip, $update_regnick_vhost,

	$chk_clone_except, $count_clones,
	
	$get_pass, $get_pass_email, 
	$set_pass, $set_email, $set_protect_level,
	
	$set_flags, $get_flags, $set_flag, $unset_flag,

	$get_user_flags, $set_user_flag, $unset_user_flag, $set_user_flag_all,

	$get_root_nick, $chk_pass, $identify, $identify_ign, $id_update, $logout, $unidentify, $unidentify_single,
	$update_lastseen, $quit_update, $protect_level,

	$register, $create_alias, $drop, $change_root, $get_nickreg, $copy_nickreg, $change_root2,

	$get_aliases, $count_aliases, $get_random_alias, $delete_alias, $delete_aliases, $get_all_access, $del_all_access,
	$change_all_access, $change_akicks, $change_founders, $change_successors, $change_svsops,
	
	$lock_user_table, $cleanup, $unlock_tables,
	$is_online,

	$get_matching_nicks,

	$cleanup_nickid, $cleanup_users, $get_expired, $get_near_expired, $set_near_expired,
	
	$get_watches, $check_watch, $set_watch, $del_watch, $chgroot_watch,
	$get_silences, $check_silence, $set_silence, $del_silence, $chgroot_silence,

	$get_seen
);

sub init() {
	$nick_check = $services::dbh->prepare("SELECT id FROM user WHERE nick=? AND online=0 AND time=?");
	$nick_create = $services::dbh->prepare("INSERT INTO user SET id=(RAND()*294967293)+1, nick=?, time=?, inval=0, ident=?, host=?, vhost=?, server=?, modes=?, gecos=?, flags=?, online=1");
	$nick_create_old = $services::dbh->prepare("UPDATE user SET ident=?, host=?, vhost=?, server=?, modes=?, gecos=?, flags=?, online=1 WHERE nick=?");
	$nick_change = $services::dbh->prepare("UPDATE user SET nick=?, time=? WHERE nick=?");
	$nick_quit = $services::dbh->prepare("UPDATE user SET online=0, quittime=UNIX_TIMESTAMP() WHERE nick=?");
	$nick_delete = $services::dbh->prepare("DELETE FROM user WHERE nick=?");
	$nick_id_delete = $services::dbh->prepare("DELETE FROM nickid WHERE id=?");
	$get_quit_empty_chans = $services::dbh->prepare("SELECT cu2.chan, COUNT(*) AS c FROM chanuser AS cu1, chanuser AS cu2 WHERE cu1.nickid=? AND cu1.chan=cu2.chan AND cu1.joined=1 AND cu2.joined=1 GROUP BY cu2.chan HAVING c=1");
	$nick_chan_delete = $services::dbh->prepare("DELETE FROM chanuser WHERE nickid=?");
	$get_hostless_nicks = $services::dbh->prepare("SELECT nick FROM user WHERE vhost='*'");

	$get_nickchg = $services::dbh->prepare("SELECT nickid FROM nickchg WHERE nick=?");
	$del_nickchg_id = $services::dbh->prepare("DELETE FROM nickchg WHERE nickid=?");
	$add_nickchg = $services::dbh->prepare("REPLACE INTO nickchg SELECT ?, id, ? FROM user WHERE nick=?");
	$reap_nickchg = $services::dbh->prepare("DELETE FROM nickchg WHERE seq<?");

	$get_user_id = $services::dbh->prepare("SELECT id FROM user WHERE nick=?");
	$get_user_nick = $services::dbh->prepare("SELECT nick FROM user WHERE id=?");
	
	$get_nick_inval = $services::dbh->prepare("SELECT nick, inval FROM user WHERE id=?");
	$inc_nick_inval = $services::dbh->prepare("UPDATE user SET inval=inval+1 WHERE id=?");

	$get_id_nicks = $services::dbh->prepare("SELECT nickalias.root FROM nickid, nickalias WHERE nickid.nick=nickalias.alias AND nickid.id=? AND nickalias.root=nickalias.alias");
	$get_using_nicks = $services::dbh->prepare("SELECT user.nick FROM user, nickid WHERE user.id=nickid.id AND nickid.nick=? AND user.online=1");
	$is_identified = $services::dbh->prepare("SELECT 1 FROM user, nickid WHERE user.id=nickid.id AND user.nick=? AND nickid.nick=?");
	$is_registered = $services::dbh->prepare("SELECT 1 FROM nickalias WHERE alias=?");
	$is_alias_of = $services::dbh->prepare("SELECT 1 FROM nickalias AS n1 LEFT JOIN nickalias AS n2 ON n1.root=n2.root WHERE n1.alias=? AND n2.alias=? LIMIT 1");

	$get_guest = $services::dbh->prepare("SELECT guest FROM user WHERE nick=?");
	$set_guest = $services::dbh->prepare("UPDATE user SET guest=? WHERE nick=?");

	$get_lock = $services::dbh->prepare("SELECT GET_LOCK(?, 10)");
	$release_lock = $services::dbh->prepare("SELECT RELEASE_LOCK(?)");

	$get_umodes = $services::dbh->prepare("SELECT modes FROM user WHERE id=?");
	$set_umodes = $services::dbh->prepare("UPDATE user SET modes=? WHERE id=?");
	
	$get_info = $services::dbh->prepare("SELECT nickreg.email, nickreg.regd, nickreg.last, nickreg.flags, nickreg.ident, nickreg.vhost, nickreg.gecos, nickreg.quit FROM nickreg, nickalias WHERE nickalias.root=nickreg.nick AND nickalias.alias=?");
	$get_host = $services::dbh->prepare("SELECT ident, host FROM user WHERE id=?");
	$get_vhost = $services::dbh->prepare("SELECT ident, vhost FROM user WHERE id=?");
	$set_ident = $services::dbh->prepare("UPDATE user SET ident=? WHERE id=?");
	$set_vhost = $services::dbh->prepare("UPDATE user SET vhost=? WHERE id=?");
	$set_ip = $services::dbh->prepare("UPDATE user SET ip=? WHERE id=?");
	$update_regnick_vhost = $services::dbh->prepare("UPDATE nickreg,nickid SET nickreg.vhost=? WHERE nickreg.nick=nickid.nick AND nickid.id=?");

	$chk_clone_except = $services::dbh->prepare("SELECT GREATEST(IF((user.ip >> (32 - sesexip.mask)) = (sesexip.ip >> (32 - sesexip.mask)), sesexip.lim, 0), IF(IF(sesexname.serv, user.server, user.host) LIKE sesexname.host, sesexname.lim, 0)) AS n FROM user, sesexip, sesexname WHERE user.id=? ORDER BY n DESC LIMIT 1");
	$count_clones = $services::dbh->prepare("SELECT COUNT(*) FROM user WHERE ip=? AND online=1");

	$get_root_nick = $services::dbh->prepare("SELECT root FROM nickalias WHERE alias=?");
	$chk_pass = $services::dbh->prepare("SELECT 1 FROM nickreg WHERE nick=? AND pass=?");
	$identify = $services::dbh->prepare("INSERT INTO nickid SELECT ?, alias FROM nickalias WHERE root=?");
	$identify_ign = $services::dbh->prepare("INSERT IGNORE INTO nickid SELECT ?, alias FROM nickalias WHERE root=?");
	$id_update = $services::dbh->prepare("UPDATE nickreg, user SET nickreg.last=UNIX_TIMESTAMP(), nickreg.ident=user.ident, nickreg.vhost=user.vhost, nickreg.gecos=user.gecos, nickreg.nearexp=0 WHERE nickreg.nick=? AND user.id=?");
	$logout = $services::dbh->prepare("DELETE FROM nickid WHERE id=?");
	$unidentify_single = $services::dbh->prepare("DELETE FROM nickid WHERE nickid.nick=?");
	$unidentify = $services::dbh->prepare("DELETE FROM nickid USING nickalias, nickid WHERE nickalias.root=? AND nickid.nick=nickalias.alias");

	$update_lastseen = $services::dbh->prepare("UPDATE nickreg,nickid SET nickreg.last=UNIX_TIMESTAMP() WHERE nickreg.nick=nickid.nick AND nickid.id=?");
	$quit_update = $services::dbh->prepare("UPDATE nickreg,nickid SET nickreg.quit=?, nickreg.last=UNIX_TIMESTAMP() WHERE nickreg.nick=nickid.nick AND nickid.id=?");
	$protect_level = $services::dbh->prepare("SELECT protect FROM nickalias WHERE alias=?");

	$set_protect_level = $services::dbh->prepare("UPDATE nickalias SET protect=? WHERE alias=?");
	$set_email = $services::dbh->prepare("UPDATE nickreg, nickalias SET nickreg.email=? WHERE nickalias.root=nickreg.nick AND nickalias.alias=?");
	$set_pass = $services::dbh->prepare("UPDATE nickreg, nickalias SET nickreg.pass=? WHERE nickalias.root=nickreg.nick AND nickalias.alias=?");
	
	$set_flags = $services::dbh->prepare("UPDATE nickreg, nickalias SET nickreg.flags=? WHERE nickalias.root=nickreg.nick AND nickalias.alias=?");
	$get_flags = $services::dbh->prepare("SELECT nickreg.flags FROM nickreg, nickalias WHERE nickalias.root=nickreg.nick AND nickalias.alias=?");
	$set_flag = $services::dbh->prepare("UPDATE nickreg, nickalias SET nickreg.flags=IF(nickreg.flags & ?, nickreg.flags, nickreg.flags ^ ?) WHERE nickalias.root=nickreg.nick AND nickalias.alias=?");
	$unset_flag = $services::dbh->prepare("UPDATE nickreg, nickalias SET nickreg.flags=IF(nickreg.flags & ?, nickreg.flags ^ ?, nickreg.flags) WHERE nickalias.root=nickreg.nick AND nickalias.alias=?");

	$get_user_flags = $services::dbh->prepare("SELECT flags FROM user WHERE id=?");
	$set_user_flag = $services::dbh->prepare("UPDATE user SET flags=IF(flags & ?, flags, flags ^ ?) WHERE id=?");
	$unset_user_flag = $services::dbh->prepare("UPDATE user SET flags=IF(flags & ?, flags ^ ?, flags) WHERE id=?");
	$set_user_flag_all = $services::dbh->prepare("UPDATE user SET flags=flags | ?");

	$get_pass = $services::dbh->prepare("SELECT nickreg.pass FROM nickreg, nickalias WHERE nickalias.root=nickreg.nick AND nickalias.alias=?");
	$get_pass_email  = $services::dbh->prepare("SELECT nickreg.pass, nickreg.email FROM nickalias, nickreg WHERE nickreg.nick=nickalias.root AND nickalias.alias=?");

	$register = $services::dbh->prepare("INSERT INTO nickreg SET nick=?, pass=?, email=?, flags=".F_HIDEMAIL().", regd=UNIX_TIMESTAMP(), last=UNIX_TIMESTAMP()");
	$create_alias = $services::dbh->prepare("INSERT INTO nickalias SET root=?, alias=?");

	$drop = $services::dbh->prepare("DELETE FROM nickreg WHERE nick=?");
	
	$get_nickreg = $services::dbh->prepare("SELECT * FROM nickreg WHERE nick=?");
	$copy_nickreg = $services::dbh->prepare("INSERT INTO nickreg SELECT ?, pass, email, regd, last, flags, ident, vhost, gecos, quit, nearexp FROM nickreg WHERE nick=?");

	$get_aliases = $services::dbh->prepare("SELECT alias FROM nickalias WHERE root=? ORDER BY alias");
	$count_aliases = $services::dbh->prepare("SELECT COUNT(*) FROM nickalias WHERE root=?");
	$get_random_alias = $services::dbh->prepare("SELECT alias FROM nickalias WHERE root=? AND alias != root LIMIT 1");
	$delete_alias = $services::dbh->prepare("DELETE FROM nickalias WHERE alias=?");
	$delete_aliases = $services::dbh->prepare("DELETE FROM nickalias WHERE root=?");
	
	$get_all_access = $services::dbh->prepare("SELECT chanacc.chan, chanacc.level, chanacc.adder, chanacc.time FROM nickalias, chanacc WHERE chanacc.nick=nickalias.root AND nickalias.alias=? ORDER BY chanacc.chan");
	$change_all_access = $services::dbh->prepare("UPDATE chanacc SET nick=? WHERE nick=?");
	$change_akicks = $services::dbh->prepare("UPDATE akick SET nick=? WHERE nick=? AND ident='' AND host=''");
	$del_all_access = $services::dbh->prepare("DELETE FROM chanacc WHERE nick=?");
	$change_founders = $services::dbh->prepare("UPDATE chanreg SET founder=? WHERE founder=?");
	$change_successors = $services::dbh->prepare("UPDATE chanreg SET successor=? WHERE successor=?");
	$change_svsops = $services::dbh->prepare("UPDATE svsop SET nick=? WHERE nick=?");
	
	$change_root = $services::dbh->prepare("UPDATE nickalias SET root=? WHERE root=?");

	$cleanup = $services::dbh->prepare("DELETE FROM nickid USING nickid LEFT JOIN user ON (nickid.id=user.id) WHERE user.id IS NULL");

	$is_online = $services::dbh->prepare("SELECT 1 FROM user WHERE nick=? AND online=1");

	$get_matching_nicks = $services::dbh->prepare("SELECT nickalias.alias, nickreg.nick, nickreg.ident, nickreg.vhost FROM nickalias, nickreg WHERE nickalias.root=nickreg.nick AND nickalias.alias LIKE ? AND nickreg.ident LIKE ? AND nickreg.vhost LIKE ? LIMIT 50");
	
	$cleanup_nickid = $services::dbh->prepare("DELETE FROM nickid, user USING nickid LEFT JOIN user ON(nickid.id=user.id) WHERE user.id IS NULL OR (user.online=0 AND quittime<?)");
	$cleanup_users = $services::dbh->prepare("DELETE FROM user WHERE online=0 AND quittime<?");
	
	$get_expired = $services::dbh->prepare("SELECT nickreg.nick, nickreg.email, nickreg.ident, nickreg.vhost FROM nickreg LEFT JOIN nickid ON(nickreg.nick=nickid.nick) LEFT JOIN svsop ON(nickreg.nick=svsop.nick) WHERE nickid.nick IS NULL AND svsop.nick IS NULL AND nickreg.last<? AND nearexp!=0 AND !(nickreg.flags & " . F_HOLD . ")");
	$get_near_expired = $services::dbh->prepare("SELECT nickreg.nick, nickreg.email FROM nickreg LEFT JOIN nickid ON(nickreg.nick=nickid.nick) LEFT JOIN svsop ON(nickreg.nick=svsop.nick) WHERE nickid.nick IS NULL AND svsop.nick IS NULL AND last<? AND nearexp=0 AND !(flags & " . F_HOLD . ")");
	$set_near_expired = $services::dbh->prepare("UPDATE nickreg SET nearexp=1, last=IF(? > last+?, ?, last) WHERE nick=?");
	
	$get_watches = $services::dbh->prepare("SELECT mask, time FROM watch WHERE nick=?");
	$check_watch = $services::dbh->prepare("SELECT 1 FROM watch WHERE nick=? AND mask=?");
	$set_watch = $services::dbh->prepare("INSERT INTO watch SET nick=?, mask=?, time=?");
	$del_watch = $services::dbh->prepare("DELETE FROM watch WHERE nick=? AND mask=?");
	$chgroot_watch = $services::dbh->prepare("UPDATE watch SET nick=? WHERE nick=?");

	$get_silences = $services::dbh->prepare("SELECT mask, time, expiry FROM silence WHERE nick=?");
	$check_silence = $services::dbh->prepare("SELECT 1 FROM silence WHERE nick=? AND mask=?");
	$set_silence = $services::dbh->prepare("INSERT INTO silence SET nick=?, mask=?, time=?, expiry=?");
	$del_silence = $services::dbh->prepare("DELETE FROM silence WHERE nick=? AND mask=?");
	$chgroot_silence = $services::dbh->prepare("UPDATE silence SET nick=? WHERE nick=?");

	$get_seen = $services::dbh->prepare("SELECT nickalias.alias, nickreg.nick, nickreg.last FROM nickreg, nickalias 
		WHERE nickalias.root=nickreg.nick AND nickalias.alias=?");
}

### NICKSERV COMMANDS ###

sub dispatch($$$) {
	my ($src, $dst, $msg) = @_;
	$msg =~ s/^\s+//;
	my @args = split(/\s+/, $msg);
	my $cmd = shift @args;

	my $user = { NICK => $src, AGENT => $nsnick };

	return if operserv::flood_check($user);

	if($cmd =~ /^help$/i) {
		help::sendhelp($nsnick, $src, 'nickserv', @args)
	}
	elsif($cmd =~ /^id(entify)?$/i) {
		if(@args == 1) {
			ns_identify($user, $src, $args[0]);
		} elsif(@args == 2) {
			ns_identify($user, $args[0], $args[1]);
		} else {
			notice($user, 'Syntax: IDENTIFY [nick] <password>');
		}
	}
	elsif($cmd =~ /^logout$/i) {
		ns_logout($user);
	}
	elsif($cmd =~ /^release$/i) {
		if(@args == 1) {
			ns_release($user, $args[0]);
		} elsif(@args == 2) {
			ns_release($user, $args[0], $args[1]);
		} else {
			notice($user, 'Syntax: RELEASE <nick> [password]');
		}
	}
	elsif($cmd =~ /^ghost$/i) {
		if(@args == 1) {
			ns_ghost($user, $args[0]);
		} elsif(@args == 2) {
			ns_ghost($user, $args[0], $args[1]);
		} else {
			notice($user, 'Syntax: GHOST <nick> [password]');
		}
	}
	elsif($cmd =~ /^register$/i) {
		if(@args == 2) {
			ns_register($user, $args[0], $args[1]);
		} else {
			notice($user, 'Syntax: REGISTER <password> <email>');
		}
	}
	elsif($cmd =~ /^(?:link|group)$/i) {
		if(@args == 2) {
			ns_link($user, $args[0], $args[1]);
		} else {
			notice($user, 'Syntax: LINK <nick> <password>');
		}
	}
	elsif($cmd =~ /^info$/i) {
		if(@args == 1) {
			ns_info($user, $args[0]);
		} else {
			notice($user, 'Syntax: INFO <nick>');
		}
	}
	elsif($cmd =~ /^set$/i) {
		if(@args == 2) {
			ns_set($user, $src, $args[0], $args[1]);
		}
		elsif(@args == 3) {
			ns_set($user, $args[0], $args[1], $args[2]);
		}
		else {
			notice($user, 'Syntax: SET <option> <value>');
			return;
		}
	}
	elsif($cmd =~ /^(drop|unlink)$/i) {
		if(@args == 1) {
			ns_unlink($user, $src, $args[0]);
		}
		elsif(@args == 2) {
			ns_unlink($user, $args[0], $args[1]);
		}
		else {
			notice($user, 'Syntax: UNLINK [nick] <password>');
		}
	}
	elsif($cmd =~ /^dropgroup$/i) {
		if(@args == 1) {
			ns_dropgroup($user, $src, $args[0]);
		}
		elsif(@args == 2) {
			ns_dropgroup($user, $args[0], $args[1]);
		}
		else {
			notice($user, 'Syntax: DROPGROUP [nick] <password>');
		}
	}
	elsif($cmd =~ /^chgroot$/i) {
		if(@args == 1) {
			ns_changeroot($user, $src, $args[0]);
		}
		elsif(@args == 2) {
			ns_changeroot($user, $args[0], $args[1]);
		}
		else {
			notice($user, 'Syntax: CHGROOT [oldroot] <newroot>');
		}
	}
	elsif($cmd =~ /^sendpass$/i) {
		if(@args == 1) {
			ns_sendpass($user, $args[0]);
		} else {
			notice($user, 'Syntax: SENDPASS <nick>');
		}
	}
	elsif($cmd =~ /^getpass$/i) {
		if(@args == 1) {
			ns_getpass($user, $args[0]);
		} else {
			notice($user, 'Syntax: GETPASS <nick>');
		}
	}
	elsif($cmd =~ /^(?:glist|links)$/i) {
		if(@args == 0) {
			ns_glist($user, $src);
		}
		elsif(@args == 1) {
			ns_glist($user, $args[0]);
		}
		else {
			notice($user, 'Syntax: GLIST [nick]');
		}
	}
	elsif($cmd =~ /^(?:alist|listchans)$/i) {
		if(@args == 0) {
			ns_alist($user, $src);
		}
		elsif(@args == 1) {
			ns_alist($user, $args[0]);
		}
		else {
			notice($user, 'Syntax: ALIST [nick]');
		}
	}
	elsif($cmd =~ /^list$/i) {
		if(@args == 1) {
			ns_list($user, $args[0]);
		} else {
			notice($user, 'Syntax: LIST <mask>');
		}
	}
	elsif($cmd =~ /^watch$/i) {
		if ($args[0] =~ /^(add|del|list)$/i) {
			ns_watch($user, $src, @args);
		}
		elsif ($args[1] =~ /^(add|del|list)$/i) {
			ns_watch($user, @args);
		}
		else {
			notice($user, 'Syntax: WATCH <ADD|DEL|LIST> [nick]');
		}
	}
	elsif($cmd =~ /^silence$/i) {
		if ($args[0] =~ /^(add|del|list)$/i) {
			ns_silence($user, $src, @args);
		}
		elsif ($args[1] =~ /^(add|del|list)$/i) {
			ns_silence($user, @args);
		}
		else {
			notice($user, 'Syntax: SILENCE <ADD|DEL|LIST> [nick]');
		}
	}
	elsif($cmd =~ /^acc(ess)?$/i) {
		if (@args == 1) {
			ns_acc($user, @args);
		}
		else {
			notice($user, 'Syntax: ACC <nick>');
		}
	}
	elsif($cmd =~ /^seen$/i) {
		if(@args >= 1) {
			ns_seen($user, @args);
		}
		else {
			notice($user, 'Syntax: SEEN <nick> [nick ...]');
		}
	}
	elsif($cmd =~ /^recover$/i) {
		if(@args == 1) {
			ns_recover($user, $args[0]);
		} elsif(@args == 2) {
			ns_recover($user, $args[0], $args[1]);
		} else {
			notice($user, 'Syntax: RECOVER <nick> [password]');
		}
	}
	else {
		notice($user, "Unrecognized command.", "For help, type: \002/msg nickserv help\002");
		log::wlog($nsnick, log::DEBUG(), "$src tried to use NickServ $msg");
	}
}

sub ns_identify($$$) {
	my ($user, $nick, $pass) = @_;
	my $src = get_user_nick($user);

	if(is_identified($user, $nick)) {
		notice($user, 'You are already identified for nick '.$nick.'.');
		return 0;
	}

	my $root = get_root_nick($nick);
	unless($root) {
		notice($user, 'Your nick is not registered.');
		return 0;
	}

	if(chk_flag($root, F_FREEZE, +1)) {
		notice($user, "This nick has been frozen and may not be used.", $err_deny);
		services::ulog($nsnick, log::INFO(), "attempted to identify to frozen nick $nick", $user);
		return;
	}

	my $uid = get_user_id($user);
	unless(chk_pass($root, $pass, $user)) {
		if(inc_nick_inval($user)) {
			notice($user, $err_pass);
		}
		services::ulog($nsnick, log::INFO(), "failed to identify to nick $nick", $user);
		return 0;
	}

	$identify->execute($uid, $root);
	$id_update->execute($root, $uid);

	#I'm unclear if this is a good idea, but it is something nice from Auspice.
	set_flag($nick, F_HOLD, 1) if(adminserv::is_ircop($user) and lc($pass) ne 'force');

	notice($user, 'You are now identified.');
	net::setumode($nsnick, $src, '+r') if lc($src) eq lc($nick);

	hostserv::hs_on($user, $root);

	nickserv::do_svssilence($user, $root);
	nickserv::do_svswatch($user, $root);

	chanserv::akick_alluser($user);
	chanserv::set_modes_allchan($user) unless chk_flag($nick, F_NEVEROP);
	
	services::ulog($nsnick, log::INFO(), "identified to nick $nick", $user);

	memoserv::notify($user, $root);

	if(enforcer_quit($nick)) {
		notice($user, 'Your nick has been released from custody.');
		return 2;
	}
	return 1;
}

sub ns_logout($) {
	my ($user) = @_;
	my $uid = get_user_id($user);
	
	$update_lastseen->execute($uid);
	$logout->execute($uid);
	notice($user, 'You are now logged out');
	services::ulog($nsnick, log::INFO(), "used NickServ LOGOUT", $user);
}

sub ns_release($$;$) {
	my ($user, $nick, $pass) = @_;

	if(chk_flag($nick, F_FREEZE, +1)) {
		notice($user, "This nick has been frozen and may not be used.", $err_deny);
		return;
	}

	unless(is_identified($user, $nick)) {
		if($pass) {
			my $s = ns_identify($user, $nick, $pass);
			return if($s == 0); #failed to identify
			if($s == 1) {
				notice($user, "Nick $nick is not being held.");
				return;
			}
		} else {
			notice($user, $err_deny);
			return;
		}
	}

	if(enforcer_quit($nick)) {
		notice($user, 'Your nick has been released from custody.');
	} else {
		notice($user, "Nick $nick is not being held.");
	}
}

sub ns_ghost($$;$) {
	my ($user, $nick, $pass) = @_;
	my $src = get_user_nick($user);

	if(chk_flag($nick, F_FREEZE, +1)) {
		notice($user, "This nick has been frozen and may not be used.", $err_deny);
		return;
	}

	unless(is_identified($user, $nick)) {
		if($pass) {
			my $s = ns_identify($user, $nick, $pass);
			return if($s == 0); #failed to identify
		} else {
			notice($user, $err_deny);
			return;
		}
	}

	if(!is_online($nick)) {
		notice($user, "\002$nick\002 is not online");
		return;
	} elsif(lc $src eq lc $nick) {
		notice($user, "I'm sorry, $src, I'm afraid I can't do that.");
		return;

	} else {
		net::irckill($nsnick, $nick, "GHOST command used by $src");
		notice($user, "Your ghost has been disconnected");
		services::ulog($nsnick, log::INFO(), "used NickServ GHOST on $nick", $user);
		#nick_delete($nick);
		return;
	}
}

sub ns_register($$$) {
	my ($user, $pass, $email) = @_;
	my $src = get_user_nick($user);
	
	if($src =~ /^guest/i) {
		notice($user, $err_deny);
		return;
	}
	
	unless(validate_email($email)) {
		notice($user, $err_email);
		return;
	}
	
	my $uid = get_user_id($user);
	
	if($create_alias->execute($src, $src)) {
		$register->execute($src, $pass, $email);
		$identify->execute($uid, $src);
		$id_update->execute($src, $uid);
		notice($user, 'You are now registered and identified.');
		net::setumode($nsnick, $src, '+r');
		services::ulog($nsnick, log::INFO(), "registered $src (email: $email)", $user);
	} else {
		notice($user, 'Your nickname has already been registered.');
	}
}

sub ns_link($$$) {
	my ($user, $nick, $pass) = @_;

	my $root = get_root_nick($nick);
	my $src = get_user_nick($user);
	my $uid = get_user_id($user);

	if($src =~ /^guest/i) {
		notice($user, $err_deny);
		return;
	}

	unless (is_registered($nick)) {
		if(is_registered($src)) {
			notice($user, "The nick \002$nick\002 is not registered. You need to change your nick to \002$nick\002 and then link to \002$src\002.");
		} else { # if neither $nick nor $src are registered
			notice($user, "You need to register your nick first. For help, type \002/ns help register");
		}
		return;
	}

	unless(chk_pass($root, $pass, $user)) {
		notice($user, $err_pass);
		return;
	}
	
	if(is_alias_of($src, $nick)) {
		notice($user, "\002$nick\002 is already linked to \002$src\002.");
		return;
	}

	if(is_registered($src)) {
		if(is_identified($user, $src)) {
			# This is disabled because it's not used much, it's easy to abuse, and it causes inconsistencies in nickid and chanacc.
			
			notice($user, "You cannot link an already registered nick. Type this and try again: \002/ns drop $src <password>");
			return;
			
			#my $sroot = get_root_nick($src);
			
			## FIXME - This may cause duplicates on access lists
			
			#$change_root->execute($root, $sroot);
			#$change_all_access->execute($root, $sroot);
			#$change_akicks->execute($root, $sroot);
			#$change_founders->execute($root, $sroot);
			#$change_successors->execute($root, $sroot);
			#$change_svsops->execute($root, $sroot);

			#$memoserv::memo_chgroot->execute($root, $sroot);
			#$memoserv::count_ignore->execute($sroot);
			#my $count = $memoserv::count_ignore->fetchrow_array;
			#$memoserv::adjust_ignore->execute($count, $sroot);
			#$memoserv::ignore_chgroot1->execute($root, $sroot);
			#$memoserv::ignore_chgroot2->execute($root, $sroot);

			#$drop->execute($sroot);

			#services::ulog($nsnick, log::INFO(), "merged nicks $sroot and $root.", $user);
		} else {
			notice($user, 'Your nickname has already been registered.');
			return;
		}
	} else {
		$create_alias->execute($root, $src);
		if(is_identified($user, $root)) {
			$identify_ign->execute($uid, $root);
			$id_update->execute($root, $uid);
		} else {
			ns_identify($user, $root, $pass);
		}
	}
	
	notice($user, "\002$src\002 is now linked to \002$root\002.");
	services::ulog($nsnick, log::INFO(), "made $src an alias of $root.", $user);

	check_identify($user);
}

sub ns_unlink($$$) {
	my ($user, $nick, $pass) = @_;
	my $uid = get_user_id($user);
	my $src = get_user_nick($user);
	
	my $root = get_root_nick($nick);
	unless(chk_pass($root, $pass, $user)) {
		notice($user, $err_pass);
		return;
	}

	if(lc $root eq lc $nick) {
		$count_aliases->execute($root);
		my ($count) = $count_aliases->fetchrow_array;
		if($count == 1) {
			ns_dropgroup_real($user, $root);
			return;
		}

		$get_random_alias->execute($root);
		my ($new) = $get_random_alias->fetchrow_array;
		ns_changeroot($user, $root, $new, 1);
		
		$root = $new;
	}
	
	unidentify_single($nick);
	delete_alias($nick);
	
	notice($user, "\002$nick\002 has been unlinked from \002$root\002.");
	services::ulog($nsnick, log::INFO(), "removed alias $nick from $root.", $user);
}

sub ns_dropgroup($$$) {
	my ($user, $nick, $pass) = @_;
	my $uid = get_user_id($user);
	my $src = get_user_nick($user);
	my $root = get_root_nick($nick);

	if(adminserv::get_svs_level($root)) {
		notice($user, "A nick with services access may not be dropped.");
		return;
	}

	unless(chk_pass($root, $pass, $user)) {
		notice($user, $err_pass);
		return;
	}

	ns_dropgroup_real($user, $root);
}

sub ns_dropgroup_real($$) {
	my ($user, $root) = @_;
	my $src = get_user_nick($user);
	
	unidentify($root, "Your nick, \002$root\002, was dropped by \002$src\002.", $src);
	dropgroup($root);
	notice($user, "Your nick(s) have been dropped.  Thanks for playing.");
	
	services::ulog($nsnick, log::INFO(), "dropped group $root.", $user);
}

sub ns_changeroot($$$;$) {
	my ($user, $old, $new, $force) = @_;

	$force or chk_identified($user, $old) or return;

	my $root = get_root_nick($old);
	
	#unless($oldroot =~ /^\Q$rootnick\E$/i) {
	#	notice($user, "\002$oldroot\002 is not a root nick.  Did you mean \002$rootnick\002?");
	#	return;
	#}

	if(lc($new) eq lc($root)) {
		notice($user, "\002$root\002 is already your root nick.");
		return;
	}
	
	unless(get_root_nick($new) eq $root) {
		notice($user, "\002$new\002 is not an alias of your nick.  Type \002/msg nickserv help link\002 for information about creating aliases.");
		return;
	}

	changeroot($root, $new);

	notice($user, "Your root nick is now \002$new\002.");
	services::ulog($nsnick, log::INFO(), "changed root $root to $new.", $user);
}

sub ns_info($$) {
	my ($user, $nick) = @_;
	
	my $root = get_root_nick($nick);
	
	$get_info->execute($root);
	my @result = $get_info->fetchrow_array;
	unless(@result) { notice($user, "The nick \002$nick\002 is not registered."); return; }
	
	my ($email, $regd, $last, $flags, $ident, $vhost, $gecos, $quit) = @result;
	my $hidemail = $flags & F_HIDEMAIL;

	my $svslev = adminserv::get_svs_level($root);
	my $protect = protect_level($nick);
	
	$get_flags->execute($nick);
	($flags) = $get_flags->fetchrow_array;

	my ($seens, $seenm) = do_seen($nick);

	my @reply = ("NickServ info for \002$nick\002:");
	#push @reply, "Alias of:        $alias" if $alias;
	my @data = (
		["Last seen address:", "$ident\@$vhost"],
		["Registered:", scalar(localtime($regd))],
	);
	
	push @data, ["Last seen:", $seenm] if $seens == 1;
	
	push @data,
		["Last real name:", $gecos],
		["Last quit:", $quit];
	
	push @data, ["Services Rank:", $adminserv::levels[$svslev]] if $svslev;
	push @data, ["E-mail:", $email] unless $hidemail;
	push @data, ["E-mail:", "$email (Hidden)"] if($hidemail and (is_identified($user, $nick) or adminserv::is_svsop($user, adminserv::S_HELP() )));

	push @reply, fmt::columnar(@data);
	pop @reply; push @reply, ' ';

	push @reply, $protect_long[$protect] if $protect;
	push @reply, "Does not accept memos." if($flags & F_NOMEMO);
	push @reply, "Cannot be added to channel access lists." if($flags & F_NOACC);
	push @reply, "Requires authorization to be added to channel access lists." if($flags & F_AUTH);
	push @reply, "Is frozen and may not be used." if($flags & F_FREEZE);
	push @reply, "Will not expire." if($flags & F_HOLD);
	push @reply, "Online now, ".$seenm.'.' if $seens == 2;

	pop @reply if($reply[-1] eq ' ');
	push @reply, ' --';
	
	notice($user, @reply);
}

sub ns_set($$$$) {
	my ($user, $target, $set, @parms) = @_;
	my $src = get_user_nick($user);
	
	unless(is_identified($user, $target) or adminserv::can_do($user, 'SERVOP')) {
		notice($user, $err_deny);
		return;
	}

	unless (
		$set =~ /^protect$/i or
		$set =~ /^e?-?mail$/i or
		$set =~ /^pass(?:w(?:or)?d)?$/i or
		$set =~ /^hidee?-?mail$/i or
		$set =~ /^nomemo$/i or
		$set =~ /^no(?:acc|op)$/i or
		$set =~ /^neverop$/i or
		$set =~ /^auth$/i or
		$set =~ /^hold$/i or
		$set =~ /^freeze$/i
	) {
		notice($user, qq{"$set" is not a valid NickServ setting.});
		return;
	}

	my ($subj, $obj);
	if($src eq $target) {
		$subj='Your';
		$obj='You';
	} else {
		$subj="\002$target\002\'s";
		$obj="\002$target\002";
	}

	if($set =~ /^protect$/i) {
		my $level;
		if($parms[0] =~ /^(?:no|off|none|false|0)$/) { $level = 0; }
		elsif($parms[0] =~ /^(?:yes|on|normal|true|1)$/) { $level = 1; }
		elsif($parms[0] =~ /^(?:high|2)$/) { $level = 2; }
		elsif($parms[0] =~ /^(?:kill|3)$/) { $level = 3; }
		else {
			notice($user, "Syntax: SET PROTECT <none|normal|high|kill>");
			return;
		}
		
		$set_protect_level->execute($level, $target);
		notice($user, "$subj protection level is now set to \002".$protect_short[$level]."\002. ".$protect_long[$level]);

		return;
	}

	if($set =~ /^e?-?mail$/i) {
		unless(@parms == 1) {
			notice($user, 'Syntax: SET EMAIL <address>');
			return;
		}
		my $email = $parms[0];
		
		unless(validate_email($email)) {
			notice($user, $err_email);
			return;
		}

		$set_email->execute($email, $target);
		notice($user, "$subj email address has been changed to \002$email\002.");
		services::ulog($nsnick, log::INFO(), "changed email of \002$target\002 to $email", $user);

		return;
	}

	if($set =~ /^pass(?:w(?:or)?d)?$/i) {
		unless(@parms == 1) {
			notice($user, 'Syntax: SET PASSWD <address>');
			return;
		}

		$set_pass->execute($parms[0], $target);
		notice($user, "$subj password has been changed.");
		services::ulog($nsnick, log::INFO(), "changed password of \002$target\002", $user);

		return;
	}

	my $val;
	if($parms[0] =~ /^(?:no|off|false|0)$/i) { $val = 0; }
	elsif($parms[0] =~ /^(?:yes|on|true|1)$/i) { $val = 1; }
	else {
		notice($user, "Please say \002on\002 or \002off\002.");
		return;
	}

	if($set =~ /^hidee?-?mail$/i) {
		set_flag($target, F_HIDEMAIL, $val);
		
		if($val) {
			notice($user, "$subj email address is now hidden.");
		} else {
			notice($user, "$subj email address is now visible.");
		}

		return;
	}

	if($set =~ /^nomemo$/i) {
		set_flag($target, F_NOMEMO, $val);

		if($val) {
			notice($user, "$subj memos will be blocked.");
		} else {
			notice($user, "$subj memos will be delivered.");
		}

		return;
	}
	
	if($set =~ /^no(?:acc|op)$/i) {
		set_flag($target, F_NOACC, $val);

		if($val) {
			notice($user, "$obj may not be added to channel access lists.");
		} else {
			notice($user, "$obj may be added to channel access lists.");
		}

		return;
	}

	if($set =~ /^neverop$/i) {
		set_flag($target, F_NEVEROP, $val);

		if($val) {
			notice($user, "$obj will not be granted status upon joining channels.");
		} else {
			notice($user, "$obj will be granted status upon joining channels.");
		}

		return;
	}

	if($set =~ /^auth$/i) {
		set_flag($target, F_AUTH, $val);

		if($val) {
			notice($user, "$obj must now authorize additions to channel access lists.");
		} else {
			notice($user, "$obj will not be asked to authorize additions to channel access lists.");
		}

		return;
	}

	if($set =~ /^hold$/i) {
		unless (adminserv::can_do($user, 'SERVOP') or
			is_identified($user, $target) and adminserv::is_ircop($user))
		{	
			notice($user, $err_deny);
			return;
		}

		set_flag($target, F_HOLD, $val);

		if($val) {
			notice($user, "\002$target\002 is now held from expiration.");
			services::ulog($nsnick, log::INFO(), "has held \002$target\002", $user);
		} else {
			notice($user, "\002$target\002 will now expire normally.");
			services::ulog($nsnick, log::INFO(), "released \002$target\002 from hold", $user);
		}

		return;
	}

	if($set =~ /^freeze$/i) {
		unless (adminserv::can_do($user, 'FREEZE') or
			is_identified($user, $target) and adminserv::is_ircop($user))
		{
			notice($user, $err_deny);
			return;
		}

		set_flag($target, F_FREEZE, $val);

		if($val) {
			notice($user, "\002$target\002 is now frozen.");
			unidentify($target, "Your nick, \002$target\002, has been frozen and may no longer be used.");
			services::ulog($nsnick, log::INFO(), "froze \002$target\002", $user);
		} else {
			notice($user, "\002$target\002 is no longer frozen.");
			services::ulog($nsnick, log::INFO(), "unfroze \002$target\002", $user);
		}

		return;
	}

	notice($user, qq{"$set" is not a valid NickServ setting.});
}

sub ns_sendpass($$) {
	my ($user, $nick) = @_;

	unless(adminserv::is_svsop($user, adminserv::S_HELP() )) {
		notice($user, $err_deny);
		return;
	}

	$get_pass_email->execute($nick);
	my ($pass, $email) = $get_pass_email->fetchrow_array;

	unless($pass and $email) {
		notice($user, "\002$nick\002 is not registered or does not have an email address.");
		return;
	}

	email::send($email, "$nsnick Password Reminder",
		"The password for the nick $nick is:\n$pass");
	
	notice($user, "Password has been sent to $email.");
	services::ulog($nsnick, log::INFO(), "used SENDPASS on $nick ($email)", $user);
}

sub ns_getpass($$) {
	my ($user, $nick) = @_;

	unless(adminserv::is_svsop($user, adminserv::S_ROOT() )) {
		notice($user, $err_deny);
		return;
	}

	my $pass = get_pass($nick);

	unless($pass) {
		notice($user, "\002$nick\002 is not registered.");
		return;
	}

	notice($user, "Password for \002$nick\002 is: \002$pass\002");
	services::ulog($nsnick, log::INFO(), "used GETPASS on $nick", $user);
}

sub ns_glist($$) {
	my ($user, $target) = @_;
	
	unless(is_identified($user, $target) or 
		adminserv::is_svsop($user, adminserv::S_HELP())
	) {
		notice($user, $err_deny);
		return;
	}
	
	my $root = get_root_nick($target);
	unless($root) {
		notice($user, "\002$target\002 is not registered.");
		return;
	}

	my @data;
	$get_aliases->execute($root);
	while(my ($alias) = $get_aliases->fetchrow_array) {
		push @data, [$alias];
	}

	notice($user,
		"The following ". $get_aliases->rows ." nicks are aliases of \002$root\002:",
		fmt::columnar(@data)
	);
}

sub ns_alist($$) {
	my ($user, $target) = @_;
	
	(adminserv::is_svsop($user, adminserv::S_HELP()) and (
		chk_registered($user, $target) or return)
	) or chk_identified($user, $target) or return;
	
	my @data;

	$get_all_access->execute($target);
	while(my ($c, $l, $a, $t) = $get_all_access->fetchrow_array) {
		push @data, [$c, $chanserv::levels[$l], ($a ? "($a)" : ''), ($t ? scalar(localtime($t)) : '')];
	}

	notice($user,
		"Access listing for \002$target\002 (".
		$get_all_access->rows." entries)",
		
		fmt::columnar(@data)
	);
}

sub ns_list($$) {
	my ($user, $mask) = @_;

	unless(adminserv::is_svsop($user, adminserv::S_HELP())) {
		notice($user, $err_deny);
		return;
	}

	my ($mnick, $mident, $mhost) = parse_mask($mask);
	
	$mnick = '%' if($mnick eq '');
	$mident = '%' if($mident eq '');
	$mhost = '%' if($mhost eq '');

	my @data;
	$get_matching_nicks->execute($mnick, $mident, $mhost);
	while(my ($rnick, $rroot, $rident, $rhost) = $get_matching_nicks->fetchrow_array) {
		push @data, [$rnick, ($rroot ne $rnick ? $rroot : ''), $rident . '@' . $rhost];
	}

	notice($user, 
		"Registered nicks matching \002$mask\002:",
		fmt::columnar(@data)
	);
}

sub ns_watch($$$;$) {
	my ($user, $target, $cmd, $mask) = @_;
	my $src = get_user_nick($user);
	
	my $root = get_root_nick($target);
	unless ($root) {
		notice($user, "\002$target\002 is not registered.");
		return;
	}
	unless(is_identified($user, $target)) {
		notice($user, $err_deny);
		return;
	}
	
	if ($cmd =~ /^add$/i) {
		if($mask =~ /\!/ or $mask =~ /\@/) {
			my ($mnick, $mident, $mhost) = parse_mask($mask);
			if ($mnick =~ /\*/) {
				notice($user, "Invalid mask: \002$mask\002", 
					'A WATCH mask cannot wildcard the nick.');
				return;
			}
		}

		$check_watch->execute($root, $mask);
		if ($check_watch->fetchrow_array) {
			notice($user, "\002$mask\002 is already in \002$target\002's watch list.");
			return;
		}

		$set_watch->execute($root, $mask, time());
		net::svswatch($nsnick, $src, "+$mask");
		notice($user, "\002$mask\002 added to \002$target\002's watch list.");
		return;
	}
	elsif ($cmd =~ /^del(ete)?$/i) {
		if($mask =~ /\!/ or $mask =~ /\@/) {
			my ($mnick, $mident, $mhost) = parse_mask($mask);
			if ($mnick =~ /\*/) {
				notice($user, "Invalid mask: \002$mask\002", 
					'A WATCH mask cannot wildcard the nick.');
			}
		}

		$check_watch->execute($root, $mask);
		unless ($check_watch->fetchrow_array) {
			notice($user, "\002$mask\002 is not in \002$target\002's watch list.");
			return;
		}
		$del_watch->execute($root, $mask);
		net::svswatch($nsnick, $src, "-$mask");
		notice($user, "\002$mask\002 removed from \002$target\002's watch list.");
	}
	elsif ($cmd =~ /^list$/i) {
		my @data;
		
		$get_watches->execute($root);
		while(my ($mask, $time) = $get_watches->fetchrow_array) {
			push @data, [$mask, scalar(localtime($time))];
		}
		
		notice($user,
			"Watch list for \002$target\002:",
			fmt::columnar(@data)
		);
	}
	else {
		notice($user, 'Syntax: WATCH <ADD|DEL|LIST> [nick]');
	}
}

sub ns_silence($$$;$$) {
	my ($user, $target, $cmd, $mask, $expiry) = @_;
	my $src = get_user_nick($user);
	
	my $root = get_root_nick($target);
	unless ($root) {
		notice($user, "\002$target\002 is not registered.");
		return;
	}
	
	unless(is_identified($user, $target)) {
		notice($user, $err_deny);
		return;
	}
	
	if ($cmd =~ /^add$/i) {
		unless($mask =~ /\!/ && $mask =~ /\@/) {
			notice($user, "Invalid mask: \002$mask\002",
				"A SILENCE mask must be in the format nick!ident\@host");
			return;
		}

		$check_silence->execute($root, $mask);
		if ($check_silence->fetchrow_array) {
			notice($user, "\002$mask\002 is already in \002$target\002's SILENCE list.");
			return;
		}
		
		if($expiry) {
			$expiry=misc::parse_time($expiry)+time();
		}
		else {
			$expiry = 0;
		};
		$set_silence->execute($root, $mask, time(), $expiry);
		net::svssilence($nsnick, $src, "+$mask");
		notice($user, "\002$mask\002 added to \002$target\002's SILENCE list.");
	}
	elsif ($cmd =~ /^del(ete)?$/i) {
		unless($mask =~ /\!/ && $mask =~ /\@/) {
			notice($user, "Invalid mask: \002$mask\002", 
				"A SILENCE mask must be in the format nick!ident\@host");
			return;
		}

		$check_silence->execute($root, $mask);
		unless ($check_silence->fetchrow_array) {
			notice($user, "\002$mask\002 is not in \002$target\002's watch list.");
			return;
		}
		$del_silence->execute($root, $mask);
		net::svssilence($nsnick, $src, "-$mask");
		notice($user, "\002$mask\002 removed from \002$target\002's SILENCE list.");
	}
	elsif ($cmd =~ /^list$/i) {
		$get_silences->execute($root);
		
		my @data;
		while(my ($mask, $time, $expiry) = $get_silences->fetchrow_array) {
			push @data, [$mask, scalar(localtime($time)),
				($expiry ? "Expires in ".misc::time_rel($expiry-time()) : "Never expires")
			];
		}
		
		notice($user,
			"Silence list for \002$target\002:",
			fmt::columnar(@data)
		);
	}
	else {
		notice($user, 'Syntax: SILENCE <ADD|DEL|LIST> [nick]');
	}
}

sub ns_acc($$) {
	my ($user, $target) = @_;
	
	unless(is_registered($target)) {
		notice($user, "$target 0", "\002$target\002 is not registered.");
		return;
	}
	
	unless(is_online($target)) {
		notice($user, "$target 1", "\002$target\002 is registered and offline.");
		return;
	}
	
	unless(is_identified({NICK => $target}, $target)) {
		notice($user, "$target 2", "\002$target\002 is online but not identified.");
		return;
	}
	
	notice($user, "$target 3", "\002$target\002 is registered and identified.");
}

sub ns_seen($@) {
	my ($user, @nicks) = @_;

	foreach my $nick (@nicks) {
		if(lc $nick eq lc $user->{AGENT}) {
			notice($user, "Oh, a wise guy, eh?");
			next;
		}
		my ($status, $msg) = do_seen($nick);
		if($status == 2) {
			notice($user, "\002$nick\002 is online now, ".$msg.'.');
		} elsif($status == 1) {
			notice($user, "\002$nick\002 was last seen ".$msg.'.');
		} else {
			notice($user, "The nick \002$nick\002 is not registered.");
		}
	}
}

sub ns_recover($$;$) {
	my ($user, $nick, $pass) = @_;
	my $src = get_user_nick($user);

	if(chk_flag($nick, F_FREEZE, +1)) {
		notice($user, "This nick has been frozen and may not be used.", $err_deny);
		return;
	}

	unless(is_identified($user, $nick)) {
		if($pass) {
			my $s = ns_identify($user, $nick, $pass);
			return if($s == 0); #failed to identify
		} else {
			notice($user, $err_deny);
			return;
		}
	}

	if(!is_online($nick)) {
		notice($user, "\002$nick\002 is not online");
		return;
	} elsif(lc $src eq lc $nick) {
		notice($user, "I'm sorry, $src, I'm afraid I can't do that.");
		return;

	} else {
		collide($nick);
		notice($user, "User claiming your nick has been collided", 
			"/msg NickServ RELEASE $nick to get it back before the one-minute timeout.");
		services::ulog($nsnick, log::INFO(), "used NickServ RECOVER on $nick", $user);
		return;
	}
}

### MISCELLANEA ###

sub do_seen($$) {
	my ($nick) = @_;
	my ($status, $msg);
	
	$get_seen->execute($nick);
	if (my ($alias, $root, $lastseen) = $get_seen->fetchrow_array) {
		$nickserv::get_using_nicks->execute($root);
		if($nickserv::get_using_nicks->rows) {
			my @usernicks;
			while (my ($usernick) = $nickserv::get_using_nicks->fetchrow_array) {
				push @usernicks, "\002$usernick\002";
			}
			$status = 2;
			$msg = "using ".(@usernicks==1 ? 'the nick ' : 'the following nicks: ').join(', ', @usernicks);
		}
		else {
			$status = 1;
			$msg = misc::time_ago($lastseen) . " ago (".scalar(localtime($lastseen)).")";
		}
	}
	else {
		$status = 0; $msg = undef();
	}

	return ($status, $msg);
}

# For a whole group:
sub unidentify($$;$) {
	my ($nick, $msg, $src) = @_;

	$nick = get_root_nick($nick);

	$get_using_nicks->execute($nick);
	while(my ($t) = $get_using_nicks->fetchrow_array) {
		net::notice($nsnick, $t, $msg) unless(lc $t eq lc $src);
		if(is_alias_of($nick, $t)) {
			net::setumode($nsnick, $t, '-r');
		}
	}

	$unidentify->execute($nick);
}

# For a single alias:
sub unidentify_single($$) {
	my ($nick, $msg) = @_;

	if(is_online($nick)) {
		net::setumode($nsnick, $nick, '-r');
	}

	$unidentify_single->execute($nick);
}

sub notice($@) {
	my $user = shift;
	
	net::notice($user->{AGENT} || $nsnick, get_user_nick($user), @_);
}

sub kill_clones($$) {
	my ($user, $ip) = @_;
	my $uid = get_user_id($user);
	my $src = get_user_nick($user);

	return 0 if $ip == 0;

	$chk_clone_except->execute($uid);
	my ($lim) = $chk_clone_except->fetchrow_array;
	return 0 if $lim == MAX_LIM();
	$lim = $services::conf{'clone-limit'} unless $lim;
	
	$count_clones->execute($ip);
	my ($c) = $count_clones->fetchrow_array;

	if($c > $lim) {
		net::irckill($nsnick, $src, "Session Limit Exceeded");
		return 1;
	}
}

sub kill_user($$) {
	my ($user, $reason) = @_;

	net::irckill(get_user_agent($user) || $config::conf{local}, get_user_nick($user), $reason);
}

sub kline_user($$$) {
	my ($user, $time, $reason) = @_;
	my $agent = get_user_agent($user);
	my ($ident, $host) = nickserv::get_host($user);

	net::kline($agent, '*', $host, $time, $reason);
}

sub parse_mask($) {
	my ($mask) = @_;
	my ($mnick, $mident, $mhost);

	$mask =~ s/([%_])/\\$1/g;
	$mask =~ tr/*?/%_/;

        $mask =~ /^(.*?)(?:\!|\@|$)/;
        $mnick = $1;

        if($mask =~ /\!(.*?)(?:\@|$)/) {
                $mident = $1;
        } else {
		$mident = '';
	}
	
        if($mask =~ /\@(.*?)$/) {
                $mhost = $1;
        } else {
		$mhost = '';
	}

	return ($mnick, $mident, $mhost);
}


sub validate_email($) {
	my ($email) = @_;
	
	if(
		$email =~ /^.+\@.+\.\w+$/ and
		$email !~ /^(abuse|postmaster|noc|security)\@/i
	) {
		return 1;
	} else {
		return 0;
	}
}

# This is mostly for logging, be careful using it for anything else
sub get_hostmask($) {
	my ($user) = @_;
	my ($ident, $host);
	my $src = get_user_nick($user);
	
	($ident, $host) = get_host($user);

	return "$src!$ident\@$host";
}

sub guestnick($) {
	my ($nick) = @_;
	
	$set_guest->execute(1, $nick);
	my $randnick = 'Guest'.int(rand(10)).int(rand(10)).int(rand(10)).int(rand(10)).int(rand(10));
	#Prevent collisions.
	while (is_online($randnick)) {
	    $randnick = 'Guest'.int(rand(10)).int(rand(10)).int(rand(10)).int(rand(10)).int(rand(10));
	}
	net::svsnick($nsnick, $nick, $randnick);

	return $randnick;
}

sub expire {
	my ($ne, $e) = ($services::conf{'nearexpire'}, $services::conf{'nickexpire'});
	
	$get_expired->execute(time() - (86400 * $e));
	while(my ($nick, $email, $ident, $vhost) = $get_expired->fetchrow_array) {
		dropgroup($nick);
		log::wlog($nsnick, log::INFO(), "$nick has expired.  Email: $email  Vhost: $ident\@$vhost");
		chanserv::drop_nick_chans($nick);
	}

	my $time = time();

	$get_near_expired->execute($time - (86400 * ($e - $ne)));
	while(my ($nick, $email) = $get_near_expired->fetchrow_array) {
		$get_aliases->execute($nick);
		my @aliases;
		while(my ($alias) = $get_aliases->fetchrow_array) {
			push @aliases, $alias;
		}
		
		my $message = "We would like to remind you that your registered nick, $nick, will expire\nin $ne days unless you sign on and identify.";
		if(@aliases > 1) {
			$message .= "\n\nThe following nicks are linked in this group:\n  " . join("\n  ", @aliases);
		}
		
		email::send($email, "$nsnick Expiration Notice", $message);
		
		$set_near_expired->execute($time, 86400 * ($e - ($ne-1)), $time - 86400 * ($e - $ne), $nick);
		log::wlog($nsnick, log::INFO(), "$nick will expire in $ne days. ($email)");
	}
}

sub do_svssilence($$) {
	my ($user, $rootnick) = @_;
	my $target = get_user_nick($user);
	
	$get_silences->execute($rootnick);
	my $count = $get_silences->rows;
	unless ($get_silences->rows) {
		$get_silences->finish;
		return;
	}
	my @silences;
	for(my $i = 1; $i <= $count; $i++) {
		my ($mask, $time, $expiry) = $get_silences->fetchrow_array;
		push @silences, "+$mask";
	}
	$get_silences->finish;
	net::svssilence($nsnick, $target, @silences);
	return;
}

sub do_svswatch($$) {
	my ($user, $rootnick) = @_;
	my $target = get_user_nick($user);
	
	$get_watches->execute($rootnick);
	my $count = $get_watches->rows;
	unless ($get_watches->rows) {
		$get_watches->finish;
		return;
	}
	my @watches;
	for(my $i = 1; $i <= $count; $i++) {
		my ($mask, $time, $expiry) = $get_watches->fetchrow_array;
		push @watches, "+$mask";
	}
	$get_watches->finish;
	net::svswatch($nsnick, $target, @watches);
	return;
}
	

### PROTECTION AND ENFORCEMENT ###

sub protect($) {
	my ($nick) = @_;

	return if chk_flag($nick, F_FREEZE());
	my $lev = protect_level($nick);
	
	net::notice($nsnick, $nick,
		"This nickname is registered and protected. If it is your",
		"nick, type \002/msg NickServ IDENTIFY <password>\002. Otherwise,",
		"please choose a different nick."
	) unless($lev==3);

	if($lev == 1) {
		warn_countdown("$nick 60");
	}
	elsif($lev==2) {
		collide($nick);
	}
	elsif($lev==3) {
		net::irckill($nsnick, $nick, "Unauthorized nick use with KILL protection enabled.");
		net::agent_connect($nick, 'enforcer', undef, '+i', 'NickServ Enforcement Agent');
		$enforcers{lc $nick} = 1;
		event::addtimer($nick, 60, __PACKAGE__, "nickserv::enforcer_quit");
	}
	
	return;
}

sub warn_countdown($) {
	my ($cookie)  = @_;
	my ($nick, $rem) = split(/ /, $cookie);
	my $user = { NICK => $nick };
	
	if (is_identified($user, $nick) or (!is_online($nick)) or (!is_registered($nick))) { return };

	if($rem == 0) {
		net::notice($nsnick, $nick,
			'Your nick is now being changed.');
		collide($nick);
	} else {
		net::notice($nsnick, $nick,
			"If you do not identify or change your nick in $rem seconds, your nick will be changed.");
		$rem -= 20;
		event::addtimer("$nick $rem", 20, __PACKAGE__, "nickserv::warn_countdown");
	}
}

sub collide($) {
	my ($nick) = @_;
	
	my $newnick = guestnick($nick);
	
	# We have to wait to do the enforcing until the nick actually changes
	$preenforce{lc $nick} = 1;
	event::addtimer($nick, 60, __PACKAGE__, "nickserv::enforcer_quit");
	return $newnick;
}

sub enforcer_quit($) {
	my ($nick) = @_;
	if($enforcers{lc $nick}) {
		undef($enforcers{lc $nick});
		net::agent_quit($nick, 'Enforcer Leaving');
		return 1;
	}
	return 0;
}

### DATABASE UTILITY FUNCTIONS ###

sub get_lock($) {
	my ($nick) = @_;
	
	$nick = lc $nick;

	$user_table++;

	if($cur_lock) {
		if($cur_lock ne $nick) {
			really_release_lock($nick);
			$user_table--;
			die("Tried to get two locks at the same time");
		}
		$cnt_lock++;
	} else {
		$cur_lock = $nick;
		$get_lock->execute($services::conf{'mysql-db'}.".user.$nick");
		$get_lock->finish;
	}
}

sub release_lock($) {
	my ($nick) = @_;
	
	$nick = lc $nick;

	$user_table--;

	if($cur_lock and $cur_lock ne $nick) {
		really_release_lock($cur_lock);
		
		die("Tried to release the wrong lock");
	}
	
	if($cnt_lock) {
		$cnt_lock--;
	} else {
		really_release_lock($nick);
	}
}

sub really_release_lock($) {
	my ($nick) = @_;

	$cnt_lock = 0;
	$release_lock->execute($services::conf{'mysql-db'}.".user.$nick");
	$release_lock->finish;
	undef $cur_lock;
}

sub set_flag($$;$) {
	my ($nick, $flag, $sign) = @_;
	$sign = 1 unless defined($sign);

	if($sign) {
		$set_flag->execute($flag, $flag, $nick);
	} else {
		$unset_flag->execute($flag, $flag, $nick);
	}
}

sub chk_flag($$;$) {
	my ($nick, $flag, $sign) = @_;
	$sign = 1 unless defined($sign);

	$get_flags->execute($nick);
	my ($flags) = $get_flags->fetchrow_array;

	return ($sign ? ($flags & $flag) : !($flags & $flag));
}

sub set_user_flag($$;$) {
	my ($user, $flag, $sign) = @_;
	my $uid = get_user_id($user);
	$sign = 1 unless defined($sign);

	if($sign) {
		$set_flag->execute($flag, $flag, $uid);
	} else {
		$unset_flag->execute($flag, $flag, $uid);
	}
}

sub chk_user_flag($$;$) {
	my ($user, $flag, $sign) = @_;
	my $uid = get_user_id($user);
	$sign = 1 unless defined($sign);

	$get_user_flags->execute($uid);
	my ($flags) = $get_user_flags->fetchrow_array;

	return ($sign ? ($flags & $flag) : !($flags & $flag));
}

sub is_agent($) {
	my ($nick) = @_;
	return (defined($net::agents{lc $nick}));
}

sub get_user_id($) {
	my ($user) = @_;
	my ($id, $n);

	return undef if(is_agent($user->{NICK}) and not $enforcers{lc $user->{NICK}});

	eval { $user->{ID} };
	if($@) {
		die("invalid get_user_id call");
	}

	if(exists($user->{ID})) { return $user->{ID}; }

	my $nick = $user->{NICK};

	# a cheat for isServer()
	if($user->{NICK} =~ /\./) {
		return $user->{ID} = undef;
	}

	if($nick eq '') {
		die("get_user_id called on empty string");
	}

	unless(synced()) {
		lock %uidcache;
		my $x = $uidcache{lc $nick};
		return $x if $x;
	}
	
	while(($n < 10 or (not synced() and !$user_table)) and !defined($id)) {
		$n++;
		$get_user_id->execute($nick);
		($id) = $get_user_id->fetchrow_array;
		unless($id) {
			$get_nickchg->execute($nick);
			($id) = $get_nickchg->fetchrow_array;
		}
	}

	unless($id) { log::wlog($nsnick, log::DEBUG(), "get_user_id($nick) failed. HDLRS: ". join(' ', @net::hdlrs)); }
	
	unless(synced()) {
		lock %uidcache;
		$uidcache{lc $nick} = $id;
	}

	return $user->{ID} = $id;
}


sub get_user_nick($) {
	my ($user) = @_;

	eval { $user->{ID} };
	if($@) {
		die("invalid get_user_nick call");
	}
	
	if(exists($user->{NICK})) { return $user->{NICK} }

	$get_user_nick->execute($user->{ID});
	my ($nick) = $get_user_nick->fetchrow_array;

	return $user->{NICK} = $nick;
}

sub get_user_agent($) {
	my ($user) = @_;

	eval { $user->{AGENT} };
	if($@) {
		die("invalid get_user_agent call");
	}
	
	if(exists($user->{AGENT})) { 
		return $user->{AGENT}
	}
	else {
		return undef;
	}
}

sub get_host($) {
	my ($user) = @_;
	my $id = get_user_id($user);
	
	$get_host->execute($id);
	my ($ident, $host) = $get_host->fetchrow_array;

	return ($ident, $host);
}

sub get_vhost($) {
	my ($user) = @_;
	my $id = get_user_id($user);
	
	$get_vhost->execute($id);
	my ($ident, $vhost) = $get_vhost->fetchrow_array;

	return ($ident, $vhost);
}

sub set_vhost($$) {
	my ($user, $vhost) = @_;
	my $id = get_user_id($user);
	
	return $set_vhost->execute($vhost, $id);
}

sub set_ident($$) {
	my ($user, $ident) = @_;
	my $id = get_user_id($user);
	
	return $set_ident->execute($ident, $id);
}

sub set_ip($$) {
	my ($user, $ip) = @_;
	my $id = get_user_id($user);

	return $set_ip->execute($ip, $id);
}

sub get_id_nicks($) {
	my ($user) = @_;
	my $id = get_user_id($user);
	my @nicks;

	$get_id_nicks->execute($id);
	while(my ($nick) = $get_id_nicks->fetchrow_array) {
		push @nicks, $nick;
	}

	return (@nicks);
}

sub get_root_nick($) {
	my ($nick) = @_;

	$get_root_nick->execute($nick);
	my ($root) = $get_root_nick->fetchrow_array;

	return $root;
}

sub drop($) {
	my ($nick) = @_;
	
	return $drop->execute($nick);
}

sub changeroot($$) {
	my ($old, $new) = @_;

	return if(lc $old eq lc $new);

	$copy_nickreg->execute($new, $old) or die("copy_nickreg($old, $new) failed");
	
	$change_root->execute($new, $old);
	$change_all_access->execute($new, $old);
	$change_akicks->execute($new, $old);
	$change_founders->execute($new, $old);
	$change_successors->execute($new, $old);
	$change_svsops->execute($new, $old);
	$chgroot_silence->execute($new, $old);
	$chgroot_watch->execute($new, $old);

	$memoserv::memo_chgroot->execute($new, $old);
	$memoserv::count_ignore->execute($old);
	my $count = $memoserv::count_ignore->fetchrow_array;
	$memoserv::adjust_ignore->execute($count, $old);
	$memoserv::ignore_chgroot1->execute($new, $old);
	$memoserv::ignore_chgroot2->execute($new, $old);

	$hostserv::vhost_chgroot->execute($new, $old);
	$drop->execute($old);
}

sub dropgroup($) {
	my ($root) = @_;
	
	$delete_aliases->execute($root);
	$del_all_access->execute($root);
	$memoserv::delete_all_memos->execute($root);
	chanserv::drop_nick_chans($root);
	$hostserv::del_vhost->execute($root);
	drop($root);
}

sub is_alias($) {
	my ($nick) = @_;
	
	return (get_root_nick($nick) eq $nick);
}

sub delete_alias($) {
	my ($nick) = @_;
	return $delete_alias->execute($nick);
}

sub delete_aliases($) {
	my ($root) = @_;
	return $delete_aliases->execute($root);
}

# How did this get here?
#
#sub get_aliases($) {
#	my ($nick) = @_;
#	
#	$get_aliases->execute($nick);
#	return $get_aliases->fetchrow_array;
#}

sub get_all_access($) {
	my ($nick) = @_;
	
	$get_all_access->execute($nick);
	return $get_all_access->fetchrow_array;
}

sub del_all_access($) {
	my ($root) = @_;
	
	return $del_all_access->execute($root);
}

sub chk_pass($$$) {
	my ($nick, $pass, $user) = @_;

	if(lc($pass) eq 'force' and adminserv::can_do($user, 'SERVOP')) {
		if(adminserv::get_best_svs_level($user) > adminserv::get_svs_level($nick)) {
			return 1;
		}
	}

	$chk_pass->execute($nick, $pass);
	return 1 if($chk_pass->fetchrow_array);
	
	return 0;
}

sub inc_nick_inval($) {
	my ($user) = @_;
	my $id = get_user_id($user);

	$inc_nick_inval->execute($id);
	$get_nick_inval->execute($id);
	my ($nick, $inval) = $get_nick_inval->fetchrow_array;
	if($inval > 3) {
		net::irckill($nsnick, $nick, 'Too many invalid passwords.');
		# unnecessary as irckill calls the quit handler.
		#nick_delete($nick);
		return 0;
	} else {
		return 1;
	}
}

sub is_identified($$) {
	my ($user, $rnick) = @_;
	my $nick = get_user_nick($user);

	$is_identified->execute($nick, $rnick);
	if($is_identified->fetchrow_array) {
		return 1;
	} else {
		return 0;
	}
}

sub chk_identified($;$) {
	my ($user, $nick) = @_;

	$nick = get_user_nick($user) unless $nick;
	
	chk_registered($user, $nick) or return 0;

	unless(is_identified($user, $nick)) {
		notice($user, $err_deny);
		return 0;
	}
	
	return 1;
}

sub is_registered($) {
	my ($nick) = @_;

	$is_registered->execute($nick);
	if($is_registered->fetchrow_array) {
		return 1;
	} else {
		return 0;
	}
}

sub chk_registered($;$) {
	my ($user, $nick) = @_;
	my $src = get_user_nick($user);
	my $what;
	
	if($nick) {
		if(lc $src eq lc $nick) {
			$what = "Your nick";
		} else {
			$what = "The nick \002$nick\002";
		}
	} else {
		$nick = get_user_nick($user) unless $nick;
		$what = "Your nick";
	}

	unless(is_registered($nick)) {
		notice($user, "$what is not registered.");
		return 0;
	}

	return 1;
}

sub is_alias_of($$) {
	$is_alias_of->execute($_[0], $_[1]);
	return ($is_alias_of->fetchrow_array ? 1 : 0);
}

sub check_identify($) {
	my ($user) = @_;
	my $nick = get_user_nick($user);
        if(is_registered($nick)) {
                if(is_identified($user, $nick)) {
                        net::setumode($nsnick, $nick, '+r');
                } else {
                        protect($nick);
                }
        }
}

sub protect_level($) {
	my ($nick) = @_;

	$protect_level->execute($nick);
	my ($lev) = $protect_level->fetchrow_array;

	return $lev;
}

sub is_online($) {
	my ($user) = @_;
	my $nick;

	if(ref($user)) {
		$nick = get_user_nick($user);
	} else {
		$nick = $user;
	}
	
	$is_online->execute($nick);
	my ($status) = $is_online->fetchrow_array;
	
	return $status;
}

sub chk_online($$) {
	my ($user, $target) = @_;

	unless(is_online($target)) {
		if(ref($target)) {
			$target = get_user_nick($target);
		}
		
		notice($user, "\002$target\002: No such user.");
		return 0;
	}
	
	return 1;
}

sub get_pass($) {
	my ($nick) = @_;
	
	$get_pass->execute($nick);
	my ($pass) = $get_pass->fetchrow_array;
	
	return $pass;
	
}

sub cleanup() {
	{
		lock %uidcache;
		undef %uidcache;
	}

	while($user_table > 0) { }
	$cleanup->execute();
}

sub cleanup_users() {
	event::addtimer('', OLD_USER_AGE(), __PACKAGE__, 'nickserv::cleanup_users');
	my $time = (time() - OLD_USER_AGE()*2);
	$cleanup_users->execute($time);
	$cleanup_nickid->execute($time);
}

sub fix_vhosts() {
	return; # XXX
	event::addtimer('fix_vhosts', 5, __PACKAGE__, 'nickserv::fix_vhosts');
	$get_hostless_nicks->execute();
	while (my ($nick) = $get_hostless_nicks->fetchrow_array) {
		net::notice($nsnick, $config::conf{'diag'}, "HOSTLESS NICK $nick");
		net::userhost($nick);
		net::userip($nick);
	}
	$get_hostless_nicks->finish();
}

sub nick_cede($) {
	my ($nick) = @_;
	my $id;

	$get_user_id->execute($nick);
	if($id = $get_user_id->fetchrow_array) {
		$nick_id_delete->execute($id);
		$nick_delete->execute($nick);
	}
}

### IRC EVENTS ###

sub nick_create {
	my ($nick, $time, $ident, $host, $vhost, $server, $modes, $gecos) = @_;
	get_lock($nick);
	
	$nick_check->execute($nick, $time);
	my ($id) = $nick_check->fetchrow_array;

	if($id) {
		$olduser{lc $nick} = 1;
		$nick_create_old->execute($ident, $host, $vhost, $server, $modes, $gecos, UF_FINISHED(), $nick);
	} else {
		nick_cede($nick);
		
		my $flags = (synced() ? UF_FINISHED() : 0);
		my $i;
		while($i < 10 and !$nick_create->execute($nick, $time, $ident, $host, $vhost, $server, $modes, $gecos, $flags)) { $i++ }
	}

	$add_nickchg->execute($net::ircline, $nick, $nick);

	release_lock($nick);

	if($vhost eq '*') {
		net::userhost($nick);
	}
	
	{
		lock %newuser;
		$newuser{lc $nick} = 1;
	}
	net::userip($nick);

	return $id;
}

sub nick_create_post($) {
	my ($nick) = @_;
	my $user = { NICK => $nick };
	my $old = $olduser{lc $nick};
	delete $olduser{lc $nick};

	operserv::do_news($nick, 'u') unless($old);

	get_lock($nick);

	check_identify($user);

	release_lock($nick);
}

sub nick_delete($$) {
	my ($nick, $quit) = @_;
	my $user = { NICK => $nick };
	
	get_lock($nick);
	
	my $id = get_user_id($user);

	$del_nickchg_id->execute($id);
	
	$quit_update->execute($quit, $id);

	$get_quit_empty_chans->execute($id);

	$nick_chan_delete->execute($id);
	$nick_quit->execute($nick);

	release_lock($nick);

	while(my ($cn) = $get_quit_empty_chans->fetchrow_array) {
		chanserv::channel_emptied({CHAN => $cn});
	}
}

sub nick_change($$$) {
	my ($old, $new, $time) = @_;

	return if(lc $old eq lc $new);

	if($preenforce{lc $old}) {
		net::agent_connect($old, 'enforcer', undef, '+i', 'NickServ Enforcement Agent');
		undef($preenforce{lc $old});
		$enforcers{lc $old} = 1;
		event::addtimer($old, 60, __PACKAGE__, "nickserv::enforcer_quit");
	}

	$user_table++;
	get_lock($old);
	nick_cede($new);
	$nick_change->execute($new, $time, $old);
	$add_nickchg->execute($net::ircline, $new, $new);
	$user_table--;
	release_lock($old);
	net::setumode($nsnick, $new, '-r');

	if($new =~ /^guest/i) {
		$get_guest->execute($new);
		if($get_guest->fetchrow_array) {
			$set_guest->execute(0, $new);
		} else {
			guestnick($new);
		}
		return;
	}
	
	check_identify({ NICK => $new });
}

sub umode($$) {
	my ($nick, $modes) = @_;
	my $user = { NICK => $nick };

	get_lock($nick);

	my $id = get_user_id($user);
	
	$get_umodes->execute($id);
	my ($omodes) = $get_umodes->fetchrow_array;
	$set_umodes->execute(modes::add($omodes, $modes, 0), $id);

	#this could be optimized
	# if -x call chghost('', $nick, realhost)
	net::userhost($nick) if ($modes =~ m/x/);

	release_lock($nick);
}

sub killhandle($$$$) {
	my ($src, $dst, $path, $reason) = @_;
	if (exists($net::agents{$dst})) {
		if (lc($dst) eq lc($nsnick)) {
			net::agent_connect('NickServ', 'services', undef, '+opqBHS', 'Nick Registration Agent');
		}
		return;
	}
	else {
		nick_delete($dst, "Killed ($src ($reason))");
	}
}

sub userip($$$) {
	my($src, $nick, $ip) = @_;
	my $user = { 'NICK' => $nick };
	my $new = $newuser{lc $nick};
	delete $newuser{lc $nick};
	#my $targetid = get_nick_id($target);
	my $iip; my @ips = split(/\./, $ip);
	for(my $i; $i < 4; $i++) {
		$iip += $ips[$i] * (2 ** ((3 - $i) * 8));
	}

	get_lock($nick);
	
	my $id = get_user_id($user);
	set_ip($user, $iip);
	my $killed = kill_clones($user, $iip);

	release_lock($nick);

	nick_create_post($nick) if(!$killed and $new);
}

sub chghost($$$) {
	my ($src, $dst, $vhost) = @_;
	my $user = { NICK => $dst };
	my $uid = get_user_id($user);
	
	#my ($cur_ident, $cur_vhost) = get_vhost($targetid);
	#if $cur_vhost eq '*' {
	$update_regnick_vhost->execute($vhost, $uid);
	$update_regnick_vhost->finish();
	#}
	
	get_lock($dst);
	
	set_vhost($user, $vhost);
	chanserv::akick_alluser($user);

	release_lock($dst);
}

sub chgident($$$) {
	my ($src, $dst, $ident) = @_;
	my $user = { NICK => $dst };
	
	set_ident($user, $ident);
	chanserv::akick_alluser($user);
}

1;
